import logging

import numpy as np
from tqdm import tqdm

from pyaerocom import const
from pyaerocom.exceptions import (
    AeronetReadError,
    MetaDataError,
    StationCoordinateError,
    VariableNotFoundError,
)
from pyaerocom.helpers import varlist_aerocom
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.mathutils import numbers_in_str
from pyaerocom.time_config import TS_TYPES
from pyaerocom.ungriddeddata import UngriddedData

logger = logging.getLogger(__name__)


class ReadAeronetBase(ReadUngriddedBase):
    """TEMPLATE: Abstract base class template for reading of Aeronet data

    Extended abstract base class, derived from low-level base class
    :class:`ReadUngriddedBase` that contains some more functionality.
    """

    __baseversion__ = "0.14_" + ReadUngriddedBase.__baseversion__

    #: column delimiter in data block of files
    COL_DELIM = ","

    #: dictionary assigning temporal resolution flags for supported datasets
    #: that are provided in a defined temporal resolution. Key is the name
    #: of the dataset and value is the corresponding ts_type
    TS_TYPES = {}

    #: dictionary specifying the file column names (values) for each Aerocom
    #: variable (keys)
    VAR_NAMES_FILE = {}

    #: Mappings for identifying variables in file (may be specified in addition
    #: to explicit variable names specified in VAR_NAMES_FILE)
    VAR_PATTERNS_FILE = {}

    #: OPTIONAL: dictionary specifying alternative column names for variables
    #: defined in :attr:`VAR_NAMES_FILE`
    ALT_VAR_NAMES_FILE = {}

    #: dictionary specifying the file column names (values) for each
    #: metadata key (cf. attributes of :class:`StationData`, e.g.
    #: 'station_name', 'longitude', 'latitude', 'altitude')
    META_NAMES_FILE = {}

    META_NAMES_FILE_ALT = ({},)

    #: name of measurement instrument
    INSTRUMENT_NAME = "sun_photometer"

    #: Default data unit that is assigned to all variables that are not
    #: specified in UNITS dictionary (cf. :attr:`UNITS`)
    DEFAULT_UNIT = "1"

    #: Variable specific units, only required for variables that deviate from
    #: :attr:`DEFAULT_UNIT` (is irrelevant for all variables that are
    #: so far supported by the implemented Aeronet products, i.e. all variables
    #: are dimensionless as specified in :attr:`DEFAULT_UNIT`)
    UNITS = {}

    IGNORE_META_KEYS = ["date", "time", "day_of_year"]

    def __init__(self, data_id=None, data_dir=None):
        super().__init__(data_id=data_id, data_dir=data_dir)

        # dictionary that contains information about the file columns
        # is written in method _update_col_index
        self._col_index = {}

        # header string referring to the content in attr. col_index. Is
        # updated whenever the former is updated (i.e. when method
        # _update_col_index is called). Can be used to check if
        # file structure changed between subsequent files so that
        # col_index is only recomputed when the file structure changes
        # and not for each file individually
        self._last_col_index_str = None
        self._last_col_order = []

        self._alt_var_cols = {}

    def _ts_type_from_data_id(self):
        if "." in self.data_id:
            ts_type = self.data_id.split(".")[-1]
            if ts_type in TS_TYPES:
                self.TS_TYPES[self.data_id] = ts_type
                return ts_type
        raise AttributeError("Failed to retrieve ts_type from data_id")

    @property
    def TS_TYPE(self):
        """Default implementation of string for temporal resolution"""
        try:
            return self.TS_TYPES[self.data_id]
        except KeyError:
            try:
                return self._ts_type_from_data_id()
            except AttributeError:
                return "undefined"

    @property
    def col_index(self):
        """Dictionary that specifies the index for each data column

        Note
        ----

        Implementation depends on the data. For instance, if the variable
        information is provided in all files (of all stations) and always in
        the same column, then this can be set as a fixed dictionary in the
        __init__ function of the implementation (see e.g. class
        :class:`ReadAeronetSunV2`).
        In other cases, it may not be ensured
        that each variable is available in all files or the column definition
        may differ between different stations. In the latter case you may
        automise the column index retrieval by providing the header names for
        each meta and data column you want to extract using the attribute
        dictionaries :attr:`META_NAMES_FILE` and :attr:`VAR_NAMES_FILE` by
        calling :func:`_update_col_index` in your implementation of
        :func:`read_file` when you reach the line that contains the header
        information.
        """
        return self._col_index

    def infer_wavelength_colname(self, colname, low=250, high=2000):
        """Get variable wavelength from column name

        Parameters
        ----------
        colname : str
            string of column name
        low : int
            lower limit of accepted value range
        high : int
            upper limit of accepted value range

        Returns
        -------
        str
            wavelength in nm as floating str

        Raises
        ------
        ValueError
            if None or more than one number is detected in variable string
        """
        nums = numbers_in_str(colname)
        if len(nums) == 1:
            if low <= int(nums[0]) <= high:
                self.logger.debug(
                    f"Succesfully extracted wavelength {nums[0]} nm from column name {colname}"
                )
                return nums[0]
        raise ValueError(f"Failed to extract wavelength from colname {colname}")

    def _update_col_index(self, col_index_str):
        """Update file column information for fast access during read_file

        Note
        ----
        If successful (no exceptions raised), then this methods overwrites the
        current column index information stored in :attr:`col_index`.

        Parameters
        ----------
        col_index_str : str
            header string of data table in files

        Returns
        -------
        dict
            dictionary containing indices (values) for each data /
            metadata key specified in ``VAR_NAMES_FILE`` and ``META_NAMES_FILE``.

        Raises
        ------
        MetaDataError
            if one of the specified meta data columns does not exist in data
        """
        cols = col_index_str.strip().split(self.COL_DELIM)
        mapping = {}
        for idx, info_str in enumerate(cols):
            if info_str in mapping:
                mapping[info_str] = "MULTI"
            else:
                mapping[info_str] = idx

        col_index = self._find_vars_name_based(mapping, cols)
        self._col_index = col_index
        self._last_col_index_str = col_index_str
        self._last_col_order = cols
        return col_index

    def _check_alternative_colnames(self, val, mapping):
        if val in self.META_NAMES_FILE_ALT:
            alt_names = self.META_NAMES_FILE_ALT[val]
            if isinstance(alt_names, str) and alt_names in mapping:
                return alt_names
            elif isinstance(alt_names, list):
                for alt_name in alt_names:
                    if alt_name in mapping:
                        return alt_name
        raise MetaDataError(
            f"Required meta-information string {val} could not be found in file header"
        )

    def _find_vars_name_based(self, mapping, cols):
        col_index = {}
        # find meta indices
        for key, val in self.META_NAMES_FILE.items():
            if val not in mapping:
                val = self._check_alternative_colnames(val, mapping)
            col_index[key] = mapping[val]
        for var, colname in self.VAR_NAMES_FILE.items():
            if colname in mapping:
                col_index[var] = mapping[colname]
            elif const.OBS_ALLOW_ALT_WAVELENGTHS:
                known = False
                if var in self.ALT_VAR_NAMES_FILE:
                    for alt_colname in self.ALT_VAR_NAMES_FILE[var]:
                        if alt_colname in mapping:
                            known = True
                            col_index[var] = mapping[alt_colname]
                if not known:
                    try:
                        idx = self._search_var_wavelength_tol(var, cols)
                        col_index[var] = idx
                    except Exception as e:
                        self.logger.info(
                            f"Failed to infer data column of variable {var} "
                            f"within wavelength tolerance range. Error:\n{repr(e)}"
                        )
        return col_index

    def _search_var_wavelength_tol(self, var, cols):
        """Find alternative variable within acceptance range"""
        var_info = const.VARS[var]
        colname = self.VAR_NAMES_FILE[var]

        wvl = var_info.wavelength_nm
        tol = var_info.obs_wavelength_tol_nm
        low, high = wvl - tol, wvl + tol
        if wvl is None:
            raise AttributeError(f"Variable {var} does not contain wavelength information")

        # variable information exists and contains wavelength info
        wvl_str = self.infer_wavelength_colname(colname)
        check_mask = colname.replace(wvl_str, "")
        if not wvl == float(wvl_str):
            raise ValueError(
                f"Wavelength mismatch between pyaerocom Variable {var} and "
                f"wavelength inferred from Aeronet column name {colname}"
            )

        # it is possible to extract wavelength from column
        # name and the extracted number corresponds to
        # the expected wavelength as inferred from
        # pyaerocom.Variable instance
        wvl_diff_min = 1e6

        # loop over header
        for i, col in enumerate(cols):
            try:
                wvl_str_col = self.infer_wavelength_colname(col)
            except Exception:
                pass
            else:
                wvl_col = float(wvl_str_col)
                if low <= wvl_col <= high:
                    mask = col.replace(wvl_str_col, "")
                    if check_mask == mask:
                        diff = abs(wvl_col - wvl)
                        if diff < wvl_diff_min:
                            wvl_diff_min = diff
                            if var not in self._alt_var_cols:
                                self._alt_var_cols[var] = []
                            if col not in self._alt_var_cols[var]:
                                self._alt_var_cols[var].append(col)
                            return i
        raise VariableNotFoundError(
            f"Did not find an alternative data column for variable {var} "
            f"within allowed wavelength tolerance range of +/- {tol} nm."
        )

    def print_all_columns(self):
        for col in self._last_col_order:
            print(col)

    def read(
        self,
        vars_to_retrieve=None,
        files=None,
        first_file=None,
        last_file=None,
        file_pattern=None,
        common_meta=None,
    ):
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded
        files : :obj:`list`, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.
        first_file : :obj:`int`, optional
            index of first file in file list to read. If None, the very first
            file in the list is used. Note: is ignored if input parameter
            `file_pattern` is specified.
        last_file : :obj:`int`, optional
            index of last file in list to read. If None, the very last file
            in the list is used. Note: is ignored if input parameter
            `file_pattern` is specified.
        file_pattern : str, optional
            string pattern for file search (cf :func:`get_file_list`)
        common_meta : dict, optional
            dictionary that contains additional metadata shared for this
            network (assigned to each metadata block of the
            :class:`UngriddedData` object that is returned)

        Returns
        -------
        UngriddedData
            data object
        """
        if common_meta is None:
            common_meta = {}
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        vars_to_retrieve = varlist_aerocom(vars_to_retrieve)
        if files is None:
            if len(self.files) == 0:
                self.get_file_list(pattern=file_pattern)
            files = self.files

        if file_pattern is None:
            if first_file is None:
                first_file = 0
            if last_file is None:
                last_file = len(files)

            files = files[first_file:last_file]

        self.read_failed = []

        data_obj = UngriddedData()
        meta_key = 0.0
        idx = 0

        # assign metadata object
        metadata = data_obj.metadata
        meta_idx = data_obj.meta_idx

        num_vars = len(vars_to_retrieve)
        num_files = len(files)
        logger.info("Reading AERONET data")
        skipped = 0
        for i in tqdm(range(num_files), disable=None):
            _file = files[i]
            try:
                station_data = self.read_file(_file, vars_to_retrieve=vars_to_retrieve)
            except AeronetReadError as e:
                self.logger.warning(f"\n{repr(e)}.")
                skipped += 1
                continue

            try:
                statmeta = station_data.get_meta()
            except StationCoordinateError as e:
                stat = station_data.station_name
                if isinstance(stat, list | np.ndarray):
                    stat = stat[0]
                logger.warning(f"\nSkipping station {stat}. Reason: {repr(e)}.\n")
                skipped += 1
                continue
            # Fill the metatdata dict
            # the location in the data set is time step dependant!
            # use the lat location here since we have to choose one location
            # in the time series plot
            meta = {}
            meta["var_info"] = {}
            meta.update(statmeta)

            meta["data_id"] = self.data_id
            meta["ts_type"] = self.TS_TYPE
            # meta['variables'] = vars_to_retrieve
            if "instrument_name" in station_data and station_data["instrument_name"] is not None:
                instr = station_data["instrument_name"]
            else:
                instr = self.INSTRUMENT_NAME
            meta["instrument_name"] = instr
            meta["data_revision"] = self.data_revision
            meta["filename"] = _file

            meta.update(**common_meta)
            # this is a list with indices of this station for each variable
            # not sure yet, if we really need that or if it speeds up things
            meta_idx[meta_key] = {}

            num_times = len(station_data["dtime"])

            # access array containing time stamps
            # TODO: check using index instead (even though not a problem here
            # since all Aerocom data files are of type timeseries)
            times = np.float64(station_data["dtime"])

            totnum = num_times * num_vars

            # check if size of data object needs to be extended
            if (idx + totnum) >= data_obj._ROWNO:
                # if totnum < data_obj._CHUNKSIZE, then the latter is used
                data_obj.add_chunk(totnum)

            for var_idx, var in enumerate(vars_to_retrieve):
                values = station_data[var]
                start = idx + var_idx * num_times
                stop = start + num_times

                # write common meta info for this station (data lon, lat and
                # altitude are set to station locations)
                data_obj._data[start:stop, data_obj._LATINDEX] = station_data["latitude"]
                data_obj._data[start:stop, data_obj._LONINDEX] = station_data["longitude"]
                data_obj._data[start:stop, data_obj._ALTITUDEINDEX] = station_data["altitude"]
                data_obj._data[start:stop, data_obj._METADATAKEYINDEX] = meta_key

                # write data to data object
                data_obj._data[start:stop, data_obj._TIMEINDEX] = times
                data_obj._data[start:stop, data_obj._DATAINDEX] = values
                data_obj._data[start:stop, data_obj._VARINDEX] = var_idx

                meta_idx[meta_key][var] = np.arange(start, stop)

                if var in station_data["var_info"]:
                    if "units" in station_data["var_info"][var]:
                        u = station_data["var_info"][var]["units"]
                    elif "unit" in station_data["var_info"][var]:
                        from pyaerocom.exceptions import MetaDataError

                        raise MetaDataError("Metadata attr unit is deprecated, please use units")
                    elif var in self.UNITS:
                        # support other units coming from the reader class
                        # needed for non 1 proxy data from aeronet
                        u = self.UNITS[var]
                    else:
                        u = self.DEFAULT_UNIT
                elif var in self.UNITS:
                    u = self.UNITS[var]
                else:
                    u = self.DEFAULT_UNIT
                meta["var_info"][var] = dict(units=u)
                if var not in data_obj.var_idx:
                    data_obj.var_idx[var] = var_idx

            idx += totnum
            metadata[meta_key] = meta
            meta_key = meta_key + 1.0

        if skipped:
            logger.warning(
                f"{skipped} out of {len(files)} files have been skipped (for "
                f"details see output)."
            )
        # shorten data_obj._data to the right number of points
        data_obj._data = data_obj._data[:idx]
        # data_obj.data_revision[self.data_id] = self.data_revision
        return data_obj



import cartopy
import numpy as np
import pytest

import pyaerocom.plot.plotcoordinates as mod
from pyaerocom.plot.mapping import init_map


@pytest.mark.parametrize(
    "args",
    [
        dict(),
        dict(legend=False),
        dict(label="bla"),
        dict(marker="^"),
        dict(ylim=(-50, -10)),
        dict(xlim=(-50, 50)),
        dict(fix_aspect=True),
        dict(title="Blaaaaa"),
        dict(gridlines=True),
        dict(ax=init_map()),
    ],
)
def test_plot_coordinates(args):
    lats = np.arange(10)
    lons = np.arange(10)
    val = mod.plot_coordinates(lons=lons, lats=lats, **args)
    assert isinstance(val, cartopy.mpl.geoaxes.GeoAxes)



from pathlib import Path

import numpy as np
import pytest

from tests.conftest import TEST_RTOL


def test_load_thessaloniki(aeronet_sda_subset_reader):
    reader = aeronet_sda_subset_reader
    files = reader.find_in_file_list("*Thessaloniki*")
    assert len(files) == 1
    assert Path(files[0]).name == "Thessaloniki.lev30"

    test_vars = ["ang4487aer", "od550aer", "od550gt1aer", "od550lt1aer"]
    data = reader.read_file(files[0], vars_to_retrieve=test_vars)

    assert all(x in data for x in test_vars)

    # more than 100 timestamps
    assert all(len(data[x]) > 100 for x in test_vars)

    assert isinstance(data["dtime"][0], np.datetime64)
    assert data["dtime"][0] == np.datetime64("2003-06-01T12:00:00"), data["dtime"][0]

    means = [np.nanmean(data[var]) for var in test_vars]
    desired = [1.4777584841303428, 0.1988665578854858, 0.036805761707404114, 0.16206080598741934]
    assert means == pytest.approx(desired, rel=TEST_RTOL)



import warnings
from warnings import warn


_cmap_lighttheme = "Blues"

DEFAULT_THEME = "light"
_COLOR_THEMES = dict(
    light=dict(
        name="light",
        cmap_map=_cmap_lighttheme,
        cmap_map_div="bwr_r",
        color_coastline="k",
    ),
    dark=dict(
        name="dark",
        cmap_map="viridis",
        cmap_map_div="PuOr_r",
        color_coastline="#e6e6e6",
    ),
)

MAP_AXES_ASPECT = 1.5
FIGSIZE_DEFAULT = (16, 10)
# text positions for annotations in scatter plots
SCAT_ANNOT_XYPOS = [
    (0.01, 0.95),
    (0.01, 0.90),
    (0.3, 0.90),
    (0.01, 0.86),
    (0.3, 0.86),
    (0.01, 0.82),
    (0.3, 0.82),
    (0.01, 0.78),
    (0.3, 0.78),
    (0.8, 0.1),
    (0.8, 0.06),
]


class ColorTheme:
    """Pyaerocom class specifiying plotting color theme

    Attributes
    ----------
    name : str
        name of color theme (e.g. "light" or "dark")
    cmap_map : str
        name of colormap or colormap for map plotting
    color_coastline : str
        coastline color for map plotting
    cmap_map_div : str
        name of diverging colormap (used in map plots when plotted value range
        crosses 0)
    cmap_map_div_shifted : bool
        boolean specifying whether center of diverging colormaps for map plots
        is supposed to be shifted to 0

    Example
    -------
    Load default color theme
    >>> theme = ColorTheme(name="dark")
    >>> print(theme)
    pyaerocom ColorTheme
    name : dark
    cmap_map : viridis
    color_coastline : #e6e6e6
    """

    def __init__(
        self,
        name="dark",
        cmap_map=None,
        color_coastline=None,
        cmap_map_div=None,
        cmap_map_div_shifted=True,
    ):
        warn(
            "matplotlib based plotting is no longer directly supported. This class may be removed in future versions.",
            DeprecationWarning,
        )

        if name not in _COLOR_THEMES:
            warn("Invalid name for color theme, using default theme")
            name = DEFAULT_THEME

        self.name = name
        self.cmap_map = cmap_map
        self.cmap_map_div = cmap_map_div
        self.cmap_map_div_shifted = cmap_map_div_shifted
        self.color_coastline = color_coastline
        self.color_map_text = "r"

        self.load_default(name)

    def load_default(self, theme_name="dark"):
        """Load default color theme

        Parameters
        ----------
        theme_name : str
            name of default theme

        Raises
        ------
        ValueError
            if ``theme_name`` is not a valid default theme

        """
        if theme_name not in _COLOR_THEMES:
            raise ValueError(
                "Default theme with name {} is not available. Choose from {}".format(
                    *_COLOR_THEMES
                )
            )
        self.from_dict(_COLOR_THEMES[theme_name])

    def from_dict(self, info_dict):
        """Import theme information from dictionary

        info_dict : dict
            dictionary containing theme settings
        """
        for k, v in info_dict.items():
            if k in self.__dict__:
                self[k] = v

    def to_dict(self):
        """Convert this object into dictionary

        Returns
        -------
        dict
            dictionary representation of this object

        """
        d = {}
        for k, v in self.__dict__.items():
            d[k] = v
        return d

    def __setitem__(self, key, value):
        if key in self.__dict__:
            self.__dict__[key] = value

    def __repr__(self):
        return self.name

    def __str__(self):
        s = "pyaerocom ColorTheme"
        for k, v in self.__dict__.items():
            s += f"\n{k} : {v}"
        return s


with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning, module="pyaerocom")
    COLOR_THEME = ColorTheme(DEFAULT_THEME)


def get_color_theme(theme_name="dark"):
    warn(
        "matplotlib based plotting is no longer directly supported. This function may be removed in future versions.",
        DeprecationWarning,
    )

    # Settings for colormap (use perceptually uniform colormaps)
    if theme_name not in _COLOR_THEMES:
        raise ValueError(f"Invalid input for theme_name, choose from {_COLOR_THEMES}")
    return ColorTheme(theme_name)



from __future__ import annotations

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pyaerocom import GriddedData
from pyaerocom.exceptions import VarNotAvailableError
from pyaerocom.io.readgridded import ReadGridded
from tests.conftest import TEST_RTOL, lustre_unavail
from tests.fixtures.tm5 import TM5_DATA_PATH

path_tm5 = str(TM5_DATA_PATH)


def init_reader():
    return ReadGridded(data_id="ECMWF_CAMS_REAN")


@pytest.fixture(scope="session")
def reader_reanalysis():
    return init_reader()


@pytest.fixture(scope="module")
def reader_tm5():
    return ReadGridded("TM5-met2010_CTRL-TEST")


@pytest.mark.parametrize(
    "input_args,mean_val",
    [
        (dict(var_name="od550aer", ts_type="monthly"), 0.0983),
        (
            dict(
                var_name="od550aer",
                ts_type="monthly",
                constraints={
                    "var_name": "od550aer",
                    "operator": "<",
                    "filter_val": 0.1,
                },
            ),
            0.2054,
        ),
        (
            dict(
                var_name="od550aer",
                ts_type="monthly",
                constraints={
                    "var_name": "od550aer",
                    "operator": ">",
                    "filter_val": 1000,
                },
            ),
            0.0983,
        ),
        (
            dict(
                var_name="od550aer",
                ts_type="monthly",
                constraints=[
                    {"var_name": "od550aer", "operator": "<", "filter_val": 0.1},
                    {"var_name": "od550aer", "operator": ">", "filter_val": 0.11},
                ],
            ),
            0.1047,
        ),
    ],
)
def test_read_var(reader_tm5: ReadGridded, input_args: dict, mean_val: float):
    data = reader_tm5.read_var(**input_args)
    assert data.cube.data.mean() == pytest.approx(mean_val, rel=1e-3)


def test_ReadGridded_class_empty():
    r = ReadGridded()
    assert r.data_id is None
    assert r.data_dir is None
    from pyaerocom.io.aerocom_browser import AerocomBrowser

    assert isinstance(r.browser, AerocomBrowser)
    with pytest.raises(AttributeError):
        r.years_avail
    assert r.vars_filename == []


def test_ReadGridded_data_dir(reader_tm5: ReadGridded):
    assert reader_tm5.data_dir == path_tm5
    assert reader_tm5._vars_2d == ["abs550aer", "od550aer"]
    assert reader_tm5._vars_3d == []


def test_ReadGridded_ts_types():
    r = ReadGridded(data_dir=path_tm5)
    assert sorted(r.ts_types) == ["daily", "monthly"]


def test_ReadGridded_read_var(reader_tm5: ReadGridded):
    r = reader_tm5
    data = r.read_var("od550aer")
    assert data.mean() == pytest.approx(0.0960723)
    with pytest.raises(VarNotAvailableError):
        r.read_var("wetso4")
    from pyaerocom.io.aux_read_cubes import add_cubes

    gridded = r.read_var("new_var", aux_vars=["abs550aer", "od550aer"], aux_fun=add_cubes)
    assert gridded.var_name == "new_var"
    assert isinstance(gridded, GriddedData)

    gridded_single_point = gridded.sel(
        latitude=0.0,
        longitude=0.0,
    )
    with pytest.raises(ValueError):
        gridded_single_point.latitude.guess_bounds()
    with pytest.raises(ValueError):
        gridded_single_point.longitude.guess_bounds()


@pytest.mark.parametrize(
    "experiments",
    [
        (["exp1"]),
        (["exp1", "exp2"]),
    ],
)
def test_ReadGridded_experiments(tmp_path: Path, experiments: list[str]):
    for exp in experiments:
        path = tmp_path / f"aerocom3_TM5-met2010_{exp}-CTRL2019_abs550aer_Column_2010_daily.nc"
        path.write_text("")

    reader = ReadGridded(data_dir=str(tmp_path))
    assert reader.experiments == experiments


@pytest.mark.parametrize(
    "vars,expected",
    [
        (["mmro3", "rho"], ["conco3", "mmro3", "rho"]),
        (["od440aer", "od870aer"], ["ang4487aer", "od440aer", "od870aer"]),
    ],
)
def test_ReadGridded_aux(tmp_path: Path, vars, expected):
    for var in vars:
        path = tmp_path / f"aerocom3_TM5-met2010_AP3-CTRL2019_{var}_Column_2010_daily.nc"
        path.write_text("")

    reader = ReadGridded(data_dir=str(tmp_path))
    for var in expected:
        assert reader.has_var(var)  # calling has_var
    assert sorted(reader.vars_provided) == sorted(expected)
    # assert r.has_var('conco3')


@pytest.mark.parametrize(
    "options,expected",
    [
        (dict(prefer_longer=True, flex_ts_type=True, ts_type="monthly"), "daily"),
        (dict(prefer_longer=False, flex_ts_type=True, ts_type="monthly"), "monthly"),
        (dict(prefer_longer=True, flex_ts_type=False, ts_type="monthly"), "monthly"),
        (dict(prefer_longer=False, flex_ts_type=False, ts_type="monthly"), "monthly"),
        (dict(prefer_longer=True, flex_ts_type=True), "daily"),
        (dict(prefer_longer=False, flex_ts_type=True), "daily"),
    ],
)
def test_ReadGridded_prefer_longer(options, expected):
    r = ReadGridded(data_dir=path_tm5)
    gridded = r.read_var("abs550aer", **options)
    assert gridded.ts_type == expected


def test_filter_query(reader_tm5: ReadGridded):
    reader_tm5.filter_query("abs550aer", ts_type="yearly", flex_ts_type=True)


@pytest.mark.parametrize(
    "years, expected",
    [
        ([2003, 2005], [2003, 2005]),
    ],
)
def test_ReadGridded_years_avail(tmp_path: Path, years, expected):
    for year in years:
        path = tmp_path / f"aerocom3_TM5-met2010_AP3-CTRL2019_od550aer_Column_{year}_daily.nc"
        path.write_text("")

    reader = ReadGridded(data_dir=str(tmp_path))
    assert sorted(reader.years_avail) == expected


def test_ReadGridded_get_var_info_from_files(reader_tm5: ReadGridded):
    info = reader_tm5.get_var_info_from_files()
    assert isinstance(info, dict)
    assert sorted(info) == ["abs550aer", "od550aer"]


# Lustre tests
START = "1-1-2003"
STOP = "31-12-2007"


@lustre_unavail
def test_file_info(reader_reanalysis: ReadGridded):
    assert isinstance(reader_reanalysis.file_info, pd.DataFrame)
    assert len(reader_reanalysis.file_info.columns) == 12


@lustre_unavail
def test_years_available(reader_reanalysis: ReadGridded):
    # reanalysis years in database will increase every year, only checking for some years in tests
    years = set(range(2003, 2024))
    assert set(reader_reanalysis.years_avail) >= years


@lustre_unavail
def test_data_dir(reader_reanalysis: ReadGridded):
    assert reader_reanalysis.data_dir.endswith(
        "aerocom/aerocom-users-database/ECMWF/ECMWF_CAMS_REAN/renamed"
    )


@lustre_unavail
def test_read_var_lustre(reader_reanalysis: ReadGridded):
    d = reader_reanalysis.read_var(var_name="od550aer", ts_type="daily", start=START, stop=STOP)

    assert isinstance(d, GriddedData)
    assert d.var_name == "od550aer"
    assert sum(d.shape) == 1826 + 161 + 320
    assert d.start == np.datetime64("2003-01-01T00:00:00.000000")
    assert d.stop == np.datetime64("2007-12-31T23:59:59.999999")
    assert d.longitude.points[[0, -1]] == pytest.approx([-180.0, 178.875], rel=TEST_RTOL)
    assert d.latitude.points[[0, -1]] == pytest.approx([90.0, -90.0], rel=TEST_RTOL)


@lustre_unavail
def test_prefer_longer(reader_reanalysis: ReadGridded):
    daily = reader_reanalysis.read_var(
        "od550aer", ts_type="monthly", flex_ts_type=True, prefer_longer=True
    )
    assert daily.ts_type == "daily"


@lustre_unavail
def test_read_vars(reader_reanalysis: ReadGridded):
    data = reader_reanalysis.read(
        ["od440aer", "od550aer", "od865aer"], ts_type="daily", start=START, stop=STOP
    )
    assert len(data) == 3
    assert all(d.shape == (1826, 161, 320) for d in data)


def test_read_climatology_file(reader_tm5: ReadGridded):
    data = reader_tm5.read_var("abs550aer", start=9999)
    assert isinstance(data, GriddedData)



import logging
import warnings
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr

from pyaerocom import const
from pyaerocom._lowlevel_helpers import BrowseDict, dict_to_str, list_to_shortstr, merge_dicts
from pyaerocom.exceptions import (
    CoordinateError,
    DataDimensionError,
    DataExtractionError,
    DataUnitError,
    MetaDataError,
    StationCoordinateError,
    TemporalResolutionError,
    VarNotAvailableError,
)
from pyaerocom.helpers import calc_climatology, isnumeric, isrange, to_datetime64
from pyaerocom.metastandards import STANDARD_META_KEYS, StationMetaData
from pyaerocom.time_resampler import TimeResampler
from pyaerocom.tstype import TsType
from pyaerocom.units_helpers import convert_unit, get_unit_conversion_fac

logger = logging.getLogger(__name__)


class StationData(StationMetaData):
    """Dict-like base class for single station data

    ToDo: write more detailed introduction

    Note
    ----
    Variable data (e.g. numpy array or pandas Series) can be directly
    assigned to the object. When  assigning variable data it is
    recommended to add variable metadata (e.g. unit, ts_type)
    in :attr:`var_info`, where key is variable name and value is dict with
    metadata entries.

    Attributes
    ----------
    dtime : list
        list / array containing time index values
    var_info : dict
        dictionary containing information about each variable
    data_err : dict
        dictionary that may be used to store uncertainty timeseries or data
        arrays associated with the different variable data.
    overlap : dict
        dictionary that may be filled to store overlapping timeseries data
        associated with one variable. This is, for instance, used in
        :func:`merge_vardata` to store overlapping data from another station.

    """

    #: List of keys that specify standard metadata attribute names. This
    #: is used e.g. in :func:`get_meta`
    STANDARD_COORD_KEYS = const.STANDARD_COORD_NAMES

    #: maximum numerical distance between coordinates associated with this
    #: station
    # _COORD_MAX_VAR = 0.1 #km
    _COORD_MAX_VAR = 5.0  # km
    STANDARD_META_KEYS = STANDARD_META_KEYS

    VALID_TS_TYPES = const.GRID_IO.TS_TYPES

    #: Keys that are ignored when accessing metadata
    PROTECTED_KEYS = [
        "dtime",
        "var_info",
        "station_coords",
        "data_err",
        "overlap",
        "numobs",
        "data_flagged",
    ]

    def __init__(self, **meta_info):
        self.dtime = []

        self.var_info = BrowseDict()

        self.station_coords = dict.fromkeys(self.STANDARD_COORD_KEYS)

        self.data_err = BrowseDict()
        self.overlap = BrowseDict()
        self.numobs = BrowseDict()
        self.data_flagged = BrowseDict()

        super().__init__(**meta_info)

    @property
    def default_vert_grid(self):
        """AeroCom default grid for vertical regridding

        For details, see :attr:`DEFAULT_VERT_GRID_DEF` in :class:`Config`

        Returns
        -------
        ndarray
            numpy array specifying default coordinates
        """
        return const.make_default_vert_grid()

    @property
    def vars_available(self):
        """Number of variables available in this data object"""
        return list(self.var_info)

    def has_var(self, var_name):
        """Checks if input variable is available in data object

        Parameters
        ----------
        var_name : str
            name of variable

        Returns
        -------
        bool
            True, if variable data is available, else False
        """
        if var_name not in self:
            return False
        if var_name not in self.var_info:
            logger.warning(
                f"Variable {var_name} exists in data but has no "
                f"metadata assigned in :attr:`var_info`"
            )
        return True

    def get_unit(self, var_name):
        """Get unit of variable data

        Parameters
        ----------
        var_name : str
            name of variable

        Returns
        -------
        str
            unit of variable

        Raises
        ------
        MetaDataError
            if unit cannot be accessed for variable
        """
        if var_name not in self.var_info:
            raise MetaDataError(f"Could not access variable metadata dict for {var_name}.")
        try:
            return str(self.var_info[var_name]["units"])
        except KeyError:
            add_str = ""
            if "unit" in self.var_info[var_name]:
                add_str = (
                    "Corresponding var_info dict contains "
                    'attr. "unit", which is deprecated, please '
                    "check corresponding reading routine. "
                )
            raise MetaDataError(
                f"Failed to access units attribute for variable {var_name}. {add_str}"
            )

    @property
    def units(self):
        """Dictionary containing units of all variables in this object"""
        ud = {}
        for var in self.var_info:
            ud[var] = self.get_unit(var)
        return ud

    def check_var_unit_aerocom(self, var_name):
        """Check if unit of input variable is AeroCom default, if not, convert

        Parameters
        ----------
        var_name : str
            name of variable

        Raises
        ------
        MetaDataError
            if unit information is not accessible for input variable name
        UnitConversionError
            if current unit cannot be converted into specified unit
            (e.g. 1 vs m-1)
        DataUnitError
            if current unit is not equal to AeroCom default and cannot
            be converted.
        """
        to_unit = const.VARS[var_name].units
        try:
            self.check_unit(var_name, to_unit)
        except Exception:
            self.convert_unit(var_name, to_unit)

    def check_unit(self, var_name, unit=None):
        """Check if variable unit corresponds to a certain unit

        Parameters
        ----------
        var_name : str
            variable name for which unit is to be checked
        unit : :obj:`str`, optional
            unit to be checked, if None, AeroCom default unit is used

        Raises
        ------
        MetaDataError
            if unit information is not accessible for input variable name
        UnitConversionError
            if current unit cannot be converted into specified unit
            (e.g. 1 vs m-1)
        DataUnitError
            if current unit is not equal to input unit but can be converted
            (e.g. 1/Mm vs 1/m)
        """
        if unit is None:
            unit = const.VARS[var_name].units
        u = self.get_unit(var_name)
        if not get_unit_conversion_fac(u, unit, var_name) == 1:
            raise DataUnitError(f"Invalid unit {u} (expected {unit})")

    def convert_unit(self, var_name, to_unit):
        """Try to convert unit of data

        Requires that unit of input variable is available in :attr:`var_info`

        Parameters
        ----------
        var_name : str
            name of variable
        to_unit : str
            new unit

        Raises
        ------
        MetaDataError
            if variable unit cannot be accessed
        UnitConversionError
            if conversion failed
        """
        unit = self.get_unit(var_name)

        data = self[var_name]
        try:
            tst = self.get_var_ts_type(var_name)
        except MetaDataError:
            tst = None
        data = convert_unit(data, from_unit=unit, to_unit=to_unit, var_name=var_name, ts_type=tst)

        self[var_name] = data
        self.var_info[var_name]["units"] = to_unit
        logger.debug(
            f"Successfully converted unit of variable {var_name} in {self.station_name} "
            f"from {unit} to {to_unit}"
        )

    def dist_other(self, other):
        """Distance to other station in km

        Parameters
        ----------
        other : StationData
            other data object

        Returns
        -------
        float
            distance between this and other station in km
        """
        from pyaerocom.geodesy import calc_distance

        cthis = self.get_station_coords()
        cother = other.get_station_coords()

        return calc_distance(
            cthis["latitude"],
            cthis["longitude"],
            cother["latitude"],
            cother["longitude"],
            cthis["altitude"],
            cother["altitude"],
        )

    def same_coords(self, other, tol_km=None):
        """Compare station coordinates of other station with this station

        Parameters
        ----------
        other : StationData
            other data object
        tol_km : float
            distance tolerance in km

        Returns
        -------
        bool
            if True, then the two object are located within the specified
            tolerance range
        """
        if tol_km is None:
            tol_km = self._COORD_MAX_VAR
        return True if self.dist_other(other) < tol_km else False

    def get_station_coords(self, force_single_value=True):
        """Return coordinates as dictionary

        This method uses the standard coordinate names defined in
        :attr:`STANDARD_COORD_KEYS` (latitude, longitude and altitude) to get
        the station coordinates. For each of these parameters tt first looks
        in :attr:`station_coords` if the parameter is defined (i.e. it is not
        None) and if not it checks if this object has an attribute that has
        this name and uses that one.

        Parameters
        ----------
        force_single_value : bool
            if True and coordinate values are lists or arrays, then they are
            collapsed to single value using mean

        Returns
        -------
        dict
            dictionary containing the retrieved coordinates

        Raises
        ------
        AttributeError
            if one of the coordinate values is invalid
        CoordinateError
            if local variation in either of the three spatial coordinates is
            found too large
        """
        output = {}
        for key in self.STANDARD_COORD_KEYS:
            # prefer explicit if defined in station_coord dictionary (e.g. altitude
            # attribute in lidar data will be an array corresponding to profile
            # altitudes)
            val = self.station_coords[key]
            if val is not None:
                if not isnumeric(val):
                    raise MetaDataError(f"Station coordinate {key} must be numeric. Got: {val}")
                output[key] = val
            else:
                val = self[key]
                if force_single_value and not isinstance(val, float | np.floating):
                    if isinstance(val, int | np.integer):
                        val = np.float64(val)
                    elif isinstance(val, list | np.ndarray):
                        # ToDo: consider tolerance to be specified in input
                        # args.
                        maxdiff = np.max(val) - np.min(val)
                        if key in ("latitude", "longitude"):
                            tol = 0.05  # ca 5km at equator
                        else:
                            tol = 100  # m altitude tolerance
                        if maxdiff > tol:
                            raise StationCoordinateError(
                                f"meas point coordinate arrays of {key} vary "
                                f"too much to reduce them to a single "
                                f"coordinate. Order of difference in {key} is "
                                f"{maxdiff} and maximum allowed is {tol}."
                            )
                        val = np.mean(val)
                    else:
                        raise AttributeError(
                            f"Invalid value encountered for coord {key}, "
                            f"need float, int, list or ndarray, got {type(val)}"
                        )
                output[key] = val
        return output

    def get_meta(
        self, force_single_value=True, quality_check=True, add_none_vals=False, add_meta_keys=None
    ):
        """Return meta-data as dictionary

        By default, only default metadata keys are considered, use parameter
        `add_meta_keys` to add additional metadata.

        Parameters
        ----------
        force_single_value : bool
            if True, then each meta value that is list or array,is converted
            to a single value.
        quality_check : bool
            if True, and coordinate values are lists or arrays, then the
            standarad deviation in the values is compared to the upper limits
            allowed in the local variation. The upper limits are specified
            in attr. ``COORD_MAX_VAR``.
        add_none_vals : bool
            Add metadata keys which have value set to None.
        add_meta_keys : str or list, optional
            Add none-standard metadata.

        Returns
        -------
        dict
            dictionary containing the retrieved meta-data

        Raises
        ------
        AttributeError
            if one of the meta entries is invalid
        MetaDataError
            in case of consistencies in meta data between individual time-stamps
        """
        if isinstance(add_meta_keys, str):
            add_meta_keys = [add_meta_keys]
        elif not isinstance(add_meta_keys, list):
            add_meta_keys = []
        meta = {}
        meta.update(self.get_station_coords(force_single_value))
        keys = [k for k in self.STANDARD_META_KEYS]
        keys.extend(add_meta_keys)
        for key in keys:
            if key not in self:
                logger.warning(f"No such key in StationData: {key}")
                continue
            elif key in self.PROTECTED_KEYS:
                # this is not metadata...
                continue
            elif key in self.STANDARD_COORD_KEYS:
                # this has been handled above
                continue
            if self[key] is None and not add_none_vals:
                logger.debug(f"No metadata available for key {key}")
                continue

            val = self[key]
            if force_single_value and isinstance(val, list | tuple | np.ndarray):
                if quality_check and not all([x == val[0] for x in val]):
                    raise MetaDataError(f"Inconsistencies in meta parameter {key}")
                val = val[0]
            meta[key] = val

        return meta

    def _check_meta_item(self, key):
        """Check if metadata item is valid

        Valid value types are dictionaries, lists, strings, numerical values
        and datetetime objects.
        """
        val = self[key]
        if val is None:
            return
        elif isinstance(val, np.ndarray):
            if val.ndim != 1:
                raise MetaDataError(
                    f"Invalid metadata entry {val} for key {key}. "
                    f"Only 1d numpy arrays are supported..."
                )
            self[key] = list(val)
        elif not isinstance(val, dict | list | str) and not isnumeric(val):
            try:
                self[key] = to_datetime64(val)
            except Exception:
                raise MetaDataError(
                    f"Invalid metadata entry {val} for key {key}. "
                    f"Only dicts, lists, strings, numerical "
                    f"values or datetime objects are supported."
                )

    def _merge_meta_item(self, key, val):
        """Merge meta item into this object

        Parameters
        ----------
        key
            key of metadata value
        val
            value to be added
        """
        current_val = self[key]
        same_type = isinstance(current_val, type(val))
        try:
            if isinstance(current_val, dict):
                if not same_type:
                    raise ValueError(f"Cannot merge meta item {key} due to type mismatch")
                elif not current_val == val:
                    self[key] = merge_dicts(current_val, val)

            elif isinstance(current_val, str):
                if not same_type:
                    if isinstance(val, list):
                        if current_val not in val:
                            newval = val.insert(0, current_val)
                        self[key] = newval
                    else:
                        raise ValueError(f"Cannot merge meta item {key} due to type mismatch")
                elif not current_val == val:
                    # both are str that may be already merged with ";" -> only
                    # add new entries
                    vals_in = [x.strip() for x in val.split(";")]

                    for item in vals_in:
                        if item not in current_val:
                            current_val += f";{item}"
                    self[key] = current_val

            elif isinstance(current_val, list):
                if not same_type:
                    val = [val]
                for item in val:
                    if item not in current_val:
                        current_val.append(item)
                self[key] = current_val

            elif isnumeric(current_val) and isnumeric(val):
                if np.isnan(current_val) and np.isnan(val):
                    self[key] = val
                elif val != current_val:
                    self[key] = [current_val, val]

            elif isinstance(val, list):
                if current_val not in val:
                    self[key] = val.insert(0, current_val)

            elif current_val != val:
                self[key] = [current_val, val]

            else:  # they shoul be the same
                assert current_val == val, (current_val, val)
        except Exception as e:
            raise MetaDataError(
                f"Failed to merge metadata entries for key {key}.\n"
                f"Value in current StationData: {current_val}\n"
                f"Value to be merged: {val}\n"
                f"Error: {repr(e)}"
            )

    def _append_meta_item(self, key, val):
        """Add a metadata item"""
        if key not in self or self[key] is None:
            self[key] = val
        else:
            self._merge_meta_item(key, val)

    def merge_meta_same_station(
        self,
        other,
        coord_tol_km=None,
        check_coords=True,
        inplace=True,
        add_meta_keys=None,
        raise_on_error=False,
    ):
        """Merge meta information from other object

        Note
        ----
        Coordinate attributes (latitude, longitude and altitude) are not
        copied as they are required to be the same in both stations. The
        latter can be checked and ensured using input argument ``check_coords``

        Parameters
        ----------
        other : StationData
            other data object
        coord_tol_km : float
            maximum distance in km between coordinates of input StationData
            object and self. Only relevant if :attr:`check_coords` is True. If
            None, then :attr:`_COORD_MAX_VAR` is used which is defined in the
            class header.
        check_coords : bool
            if True, the coordinates are compared and checked if they are lying
            within a certain distance to each other (cf. :attr:`coord_tol_km`).
        inplace : bool
            if True, the metadata from the other station is added to the
            metadata of this station, else, a new station is returned with the
            merged attributes.
        add_meta_keys : str or list, optional
            additional non-standard metadata keys that are supposed to be
            considered for merging.
        raise_on_error : bool
            if True, then an Exception will be raised in case one of the
            metadata items cannot be merged, which is most often due to
            unresolvable type differences of metadata values between the two
            objects

        """
        if add_meta_keys is None:
            add_meta_keys = []

        elif isinstance(add_meta_keys, str):
            add_meta_keys = [add_meta_keys]

        if not inplace:
            obj = self.copy()
        else:
            obj = self

        if check_coords:
            if coord_tol_km is None:
                coord_tol_km = self._COORD_MAX_VAR
            try:
                if not self.same_coords(other, coord_tol_km):
                    raise CoordinateError(
                        f"Station coordinates of {self.station_id} and {other.station_id} for species {self.vars_available} differ by more than {coord_tol_km} km."
                    )
            except MetaDataError:  #
                pass

        keys = [k for k in self.STANDARD_META_KEYS]
        keys.extend(add_meta_keys)
        for key in keys:
            if key in self.STANDARD_COORD_KEYS:
                if self[key] is None and other[key] is not None:
                    self[key] = other[key]
            elif key in self.PROTECTED_KEYS:
                continue
            elif key in other and other[key] is not None:
                try:
                    self._check_meta_item(key)
                    other._check_meta_item(key)

                    obj._append_meta_item(key, other[key])
                except MetaDataError as e:
                    obj[key] = "N/A_FAILED_TO_MERGE"
                    msg = f"Failed to merge meta item {key}. Reason:{repr(e)}"
                    if raise_on_error:
                        raise MetaDataError(msg)
                    else:
                        logger.warning(msg)

        return obj

    def merge_varinfo(self, other, var_name):
        """Merge variable specific meta information from other object

        Parameters
        ----------
        other : StationData
            other data object
        var_name : str
            variable name for which info is to be merged (needs to be both
            available in this object and the provided other object)
        """
        if var_name not in self.var_info or var_name not in other.var_info:
            raise MetaDataError(f"No variable meta information available for {var_name}")

        info_this = self.var_info[var_name]
        info_other = other.var_info[var_name]
        for key, val in info_other.items():
            if key not in info_this or info_this[key] is None:
                info_this[key] = val
            else:
                if isinstance(info_this[key], str):
                    if not isinstance(val, str):
                        raise ValueError(f"Cannot merge meta item {key} due to type mismatch")
                    vals = [x.strip() for x in info_this[key].split(";")]
                    vals_in = [x.strip() for x in val.split(";")]

                    for _val in vals_in:
                        if _val not in vals:
                            info_this[key] = info_this[key] + f";{_val}"
                else:
                    if isinstance(val, list | np.ndarray):
                        if len(val) == 0 or info_this[key] == val:
                            continue

                        raise ValueError(
                            "Cannot append metadata value that is "
                            "already a list or numpy array due to "
                            "potential ambiguities"
                        )
                    if isinstance(info_this[key], list):
                        if val not in info_this[key]:
                            info_this[key].append(val)
                    else:
                        if not info_this[key] == val:
                            info_this[key] = [info_this[key], val]
        return self

    def check_if_3d(self, var_name):
        """Checks if altitude data is available in this object"""
        if "altitude" in self:
            val = self["altitude"]
            if isnumeric(val):  # is numerical value
                return False
            # unique altitude values
            uvals = np.unique(val)
            if len(uvals) == 1:  # only one value in altitude array (NOT 3D)
                return False
            elif (
                len(uvals[~np.isnan(uvals)]) == 1
            ):  # only 2 unique values in altitude array but one is NaN
                return False
            return True
        return False

    def _check_ts_types_for_merge(self, other, var_name):
        ts_type = self.get_var_ts_type(var_name)
        ts_type1 = other.get_var_ts_type(var_name)
        if ts_type != ts_type1:
            # make sure each variable in the object has explicitely ts_type
            # assigned (rather than global specification)

            self._update_var_timeinfo()
            other._update_var_timeinfo()

            from pyaerocom.helpers import get_lowest_resolution

            ts_type = get_lowest_resolution(ts_type, ts_type1)
        return ts_type

    def _update_var_timeinfo(self):
        for var, info in self.var_info.items():
            data = self[var]
            if not isinstance(data, pd.Series):
                try:
                    self[var] = pd.Series(data, self.dtime)
                except Exception as e:
                    raise Exception(f"Unexpected error: {repr(e)}.\nPlease debug...")
            if "ts_type" not in info or info["ts_type"] is None:
                if self.ts_type not in const.GRID_IO.TS_TYPES:
                    raise ValueError(f"Cannot identify ts_type for var {var} in {self}")
                info["ts_type"] = self.ts_type
        self.ts_type = None

    def _merge_vardata_2d(self, other, var_name, resample_how=None, min_num_obs=None):
        """Merge 2D variable data (for details see :func:`merge_vardata`)"""
        ts_type = self._check_ts_types_for_merge(other, var_name)

        s0 = self.resample_time(
            var_name, ts_type=ts_type, how=resample_how, min_num_obs=min_num_obs, inplace=True
        )[var_name].dropna()
        s1 = other.resample_time(
            var_name, ts_type=ts_type, how=resample_how, min_num_obs=min_num_obs, inplace=True
        )[var_name].dropna()

        info = other.var_info[var_name]
        removed = None
        if "overlap" in info and info["overlap"]:
            raise NotImplementedError("Coming soon...")

        if len(s1) > 0:  # there is data
            overlap = s0.index.intersection(s1.index)
            try:
                if len(overlap) > 0:
                    removed = s1[overlap]
                    # NOTE JGLISS: updated on 8.5.2020, cf. issue #106
                    # s1 = s1.drop(index=overlap, inplace=True)
                    s1.drop(index=overlap, inplace=True)
                # compute merged time series
                if len(s1) > 0:
                    s0 = pd.concat([s0, s1], verify_integrity=True)

                # sort the concatenated series based on timestamps
                s0.sort_index(inplace=True)
                self.merge_varinfo(other, var_name)
            except KeyError:
                logger.warning(
                    f"failed to merge {var_name} data from 2 StationData "
                    f"objects for station {self.station_name}. Ignoring 2nd "
                    f"data object."
                )

        # assign merged time series (overwrites previous one)
        self[var_name] = s0
        self.dtime = s0.index.values

        if removed is not None:
            if var_name in self.overlap:
                self.overlap[var_name] = pd.concat([self.overlap[var_name], removed])
                self.overlap[var_name].sort_index(inplace=True)
            else:
                self.overlap[var_name] = removed

        return self

    def merge_vardata(self, other, var_name, **kwargs):
        """Merge variable data from other object into this object

        Note
        ----
        This merges also the information about this variable in the dict
        :attr:`var_info`. It is required, that variable meta-info is
        specified in both StationData objects.

        Note
        ----
        This method removes NaN's from the existing time series in the data
        objects. In order to fill up the time-series with NaNs again after
        merging, call :func:`insert_nans_timeseries`

        Parameters
        ----------
        other : StationData
            other data object
        var_name : str
            variable name for which info is to be merged (needs to be both
            available in this object and the provided other object)
        kwargs
            keyword args passed on to :func:`_merge_vardata_2d`

        Returns
        -------
        StationData
            this object merged with other object
        """
        if var_name not in self:
            raise VarNotAvailableError(
                f"StationData object does not contain data for variable {var_name}"
            )
        elif var_name not in other:
            raise VarNotAvailableError(
                "Input StationData object does not contain data for variable {var_name}"
            )
        elif var_name not in self.var_info:
            raise MetaDataError(
                f"For merging of {var_name} data, variable specific meta "
                f"data needs to be available in var_info dict"
            )
        elif var_name not in other.var_info:
            raise MetaDataError(
                f"For merging of {var_name} data, variable specific meta "
                f"data needs to be available in var_info dict"
            )

        if self.get_unit(var_name) != other.get_unit(var_name):
            self.check_var_unit_aerocom(var_name)
            other.check_var_unit_aerocom(var_name)

        if self.check_if_3d(var_name):
            raise NotImplementedError("Coming soon...")
            # return self._merge_vardata_3d(other, var_name)
        else:
            return self._merge_vardata_2d(other, var_name, **kwargs)

    def merge_other(self, other, var_name, add_meta_keys=None, **kwargs):
        """Merge other station data object

        Todo
        ----
        Should be independent of variable, i.e. it should be able to merge all
        data that is in the other object into this, even if this object does
        not contain that variable yet.

        Parameters
        ----------
        other : StationData
            other data object
        var_name : str
            variable name for which info is to be merged (needs to be both
            available in this object and the provided other object)
        add_meta_keys : str or list, optional
            additional non-standard metadata keys that are supposed to be
            considered for merging.
        kwargs
            keyword args passed on to :func:`merge_vardata` (e.g time
            resampling settings)

        Returns
        -------
        StationData
            this object that has merged the other station
        """
        self.merge_vardata(other, var_name, **kwargs)
        self.merge_meta_same_station(other, add_meta_keys=add_meta_keys)

        return self

    def check_dtime(self):
        """Checks if dtime attribute is array or list"""
        if not any([isinstance(self.dtime, x) for x in [list, np.ndarray]]):
            raise TypeError(f"dtime attribute is not iterable: {self.dtime}")
        elif not len(self.dtime) > 0:
            raise AttributeError("No timestamps available")

    def get_var_ts_type(self, var_name, try_infer=True):
        """Get ts_type for a certain variable

        Note
        ----
        Converts to ts_type string if assigned ts_type is in pandas format

        Parameters
        ----------
        var_name : str
            data variable name for which the ts_type is supposed to be
            retrieved
        try_infer : bool
            if ts_type is not available, try inferring it from data

        Returns
        -------
        str
            the corresponding data time resolution

        Raises
        ------
        MetaDataError
            if no metadata is available for this variable (e.g. if ``var_name``
            cannot be found in :attr:`var_info`)
        """
        # make sure there exists a var_info dict for this variable
        if var_name not in self.var_info:
            self.var_info[var_name] = {}

        # use variable specific entry if available
        if "ts_type" in self.var_info[var_name]:
            return TsType(self.var_info[var_name]["ts_type"]).val
        elif isinstance(self.ts_type, str):
            # ensures validity and corrects for pandas strings
            ts_type = TsType(self.ts_type).val
            self.var_info[var_name]["ts_type"] = ts_type
            return ts_type

        if try_infer:
            logger.warning(
                f"Trying to infer ts_type in StationData {self.station_name} "
                f"for variable {var_name}"
            )
            from pyaerocom.helpers import infer_time_resolution

            try:
                s = self._to_ts_helper(var_name)
                ts_type = infer_time_resolution(s.index)
                self.var_info[var_name]["ts_type"] = ts_type
                return ts_type
            except Exception:
                pass  # Raise standard error
        raise MetaDataError(f"Could not access ts_type for {var_name}")

    def remove_outliers(self, var_name, low=None, high=None, check_unit=True):
        """Remove outliers from one of the variable timeseries

        Parameters
        ----------
        var_name : str
            variable name
        low : float
            lower end of valid range for input variable. If None, then the
            corresponding value from the default settings for this variable
            are used (cf. minimum attribute of `available variables
            <https://pyaerocom.met.no/config_files.html#variables>`__)
        high : float
            upper end of valid range for input variable. If None, then the
            corresponding value from the default settings for this variable
            are used (cf. maximum attribute of `available variables
            <https://pyaerocom.met.no/config_files.html#variables>`__)
        check_unit : bool
            if True, the unit of the data is checked against AeroCom default
        """
        if any([x is None for x in (low, high)]):
            info = const.VARS[var_name]
            if check_unit:
                try:
                    self.check_unit(var_name)
                except DataUnitError:
                    self.convert_unit(var_name, to_unit=info.units)
            if low is None:
                low = info.minimum
                logger.info(f"Setting {var_name} outlier lower lim: {low:.2f}")
            if high is None:
                high = info.maximum
                logger.info(f"Setting {var_name} outlier upper lim: {high:.2f}")

        d = self[var_name]
        invalid_mask = np.logical_or(d < low, d > high)
        d[invalid_mask] = np.nan
        self[var_name] = d

    def calc_climatology(
        self,
        var_name,
        start=None,
        stop=None,
        min_num_obs=None,
        clim_mincount=None,
        clim_freq=None,
        set_year=None,
        resample_how=None,
    ):
        """Calculate climatological timeseries for input variable


        Parameters
        ----------
        var_name : str
            name of data variable
        start
            start time of data used to compute climatology
        stop
            start time of data used to compute climatology
        min_num_obs : dict or int, optional
            minimum number of observations required per period (when
            downsampling). For details see
            :func:`pyaerocom.time_resampler.TimeResampler.resample`)
        clim_micount : int, optional
            minimum number of of monthly values required per month of
            climatology
        set_year : int, optional
            if specified, the output data will be assigned the input year. Else
            the middle year of the climatological interval is used.
        resample_how : str
            how should the resampled data be averaged (e.g. mean, median)
        **kwargs
            Additional keyword args passed to
            :func:`pyaerocom.time_resampler.TimeResampler.resample`

        Returns
        -------
        StationData
            new instance of StationData containing climatological data
        """
        if clim_freq is None:
            clim_freq = const.CLIM_FREQ

        if resample_how is None:
            resample_how = const.CLIM_RESAMPLE_HOW

        ts_type = TsType(self.get_var_ts_type(var_name))

        monthly = TsType("monthly")
        if ts_type < monthly:
            raise TemporalResolutionError(
                f"Cannot compute climatology, {var_name} data "
                f"needs to be in monthly resolution or higher (is: {ts_type})"
            )
        if ts_type < TsType(
            clim_freq
        ):  # current resolution is lower than input climatological freq
            supported = list(const.CLIM_MIN_COUNT)
            if str(ts_type) in supported:
                clim_freq = str(ts_type)
            else:  # use monthly
                clim_freq = "monthly"

        data = self.resample_time(
            var_name, ts_type=clim_freq, how=resample_how, min_num_obs=min_num_obs, inplace=False
        )
        ts = data.to_timeseries(var_name)

        if start is None:
            start = const.CLIM_START
        if stop is None:
            stop = const.CLIM_STOP

        if clim_mincount is None:
            clim_mincount = const.CLIM_MIN_COUNT[clim_freq]

        clim = calc_climatology(
            ts, start, stop, min_count=clim_mincount, set_year=set_year, resample_how=resample_how
        )

        new = StationData()
        try:
            new.update(self.get_meta())
        except MetaDataError:
            new.update(self.get_meta(force_single_value=False))

        new[var_name] = clim["data"]
        vi = {}
        if var_name in self.var_info:
            vi.update(self.var_info[var_name])

        new.var_info[var_name] = vi
        new.var_info[var_name]["ts_type"] = "monthly"
        new.var_info[var_name]["ts_type_src"] = ts_type.base
        new.var_info[var_name]["is_climatology"] = True
        new.var_info[var_name]["clim_start"] = start
        new.var_info[var_name]["clim_stop"] = stop
        new.var_info[var_name]["clim_freq"] = clim_freq
        new.var_info[var_name]["clim_how"] = resample_how
        new.var_info[var_name]["clim_mincount"] = clim_mincount
        new.data_err[var_name] = clim["std"]
        new.numobs[var_name] = clim["numobs"]
        return new

    def resample_time(
        self, var_name, ts_type, how=None, min_num_obs=None, inplace=False, **kwargs
    ):
        """Resample one of the time-series in this object

        Parameters
        ----------
        var_name : str
            name of data variable
        ts_type : str
            new frequency string (can be pyaerocom ts_type or valid pandas
            frequency string)
        how : str
            how should the resampled data be averaged (e.g. mean, median)
        min_num_obs : dict or int, optional
            minimum number of observations required per period (when
            downsampling). For details see
            :func:`pyaerocom.time_resampler.TimeResampler.resample`)
        inplace : bool
            if True, then the current data object stored in self, will be
            overwritten with the resampled time-series
        **kwargs
            Additional keyword args passed to
            :func:`pyaerocom.time_resampler.TimeResampler.resample`

        Returns
        -------
        StationData
            with resampled variable timeseries
        """
        if inplace:
            outdata = self
        else:
            outdata = self.copy()
        if var_name not in outdata:
            raise KeyError(f"Variable {var_name} does not exist")

        to_ts_type = TsType(ts_type)  # make sure to use AeroCom ts_type

        try:
            from_ts_type = TsType(outdata.get_var_ts_type(var_name))
        except (MetaDataError, TemporalResolutionError):
            from_ts_type = None
            logger.warning(
                f"Failed to access current temporal resolution of {var_name} data "
                f"in StationData {outdata.station_name}. "
                f"No resampling constraints will be applied"
            )

        data = outdata[var_name]

        if not isinstance(data, pd.Series | xr.DataArray):
            data = outdata.to_timeseries(var_name)
        resampler = TimeResampler(data)
        new = resampler.resample(
            to_ts_type=to_ts_type,
            from_ts_type=from_ts_type,
            how=how,
            min_num_obs=min_num_obs,
            **kwargs,
        )

        outdata[var_name] = new
        outdata.var_info[var_name]["ts_type"] = to_ts_type.val
        outdata.var_info[var_name].update(resampler.last_setup)
        # there is other variables that are not resampled
        if len(outdata.var_info) > 1 and outdata.ts_type is not None:
            _tt = outdata.ts_type
            outdata.ts_type = None
            outdata.dtime = None
            for var, info in outdata.var_info.items():
                if not var == var_name:
                    info["ts_type"] = _tt
        else:  # no other variables, update global class attributes
            outdata.ts_type = to_ts_type.val
            outdata.dtime = new.index.values

        return outdata

    def resample_timeseries(self, var_name, **kwargs):
        """Wrapper for :func:`resample_time` (for backwards compatibility)

        Note
        ----
        For backwards compatibility, this method will return a pandas Series
        instead of the actual StationData object
        """
        warnings.warn(
            "This method was renamed to resample_time as a means "
            "of harmonisation with GriddedData and ColocatedData",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.resample_time(var_name, **kwargs)[var_name]

    def remove_variable(self, var_name):
        """Remove variable data

        Parameters
        ----------
        var_name : str
            name of variable that is to be removed

        Returns
        -------
        StationData
            current instance of this object, with data removed

        Raises
        ------
        VarNotAvailableError
            if the input variable is not available in this object
        """
        if not self.has_var(var_name):
            raise VarNotAvailableError(f"No such variable in StationData: {var_name}")
        self.pop(var_name)
        if var_name in self.var_info:
            self.var_info.pop(var_name)
        return self

    def insert_nans_timeseries(self, var_name):
        """Fill up missing values with NaNs in an existing time series

        Note
        ----
        This method does a resample of the data onto a regular grid. Thus, if
        the input ``ts_type`` is different from the actual current ``ts_type``
        of the data, this method will not only insert NaNs but at the same.

        Parameters
        ---------
        var_name : str
            variable name
        inplace : bool
            if True, the actual data in this object will be overwritten with
            the new data that contains NaNs

        Returns
        -------
        StationData
            the modified station data object

        """
        ts_type = self.get_var_ts_type(var_name)

        self.resample_time(var_name, ts_type, inplace=True)

        return self

    def _to_ts_helper(self, var_name):
        """Convert data internally to pandas.Series if it is not stored as such

        Parameters
        ----------
        var_name : str
            variable name of data

        Returns
        -------
        pandas.Series
            data as timeseries
        """
        data = self[var_name]
        if isinstance(data, pd.Series):
            return data

        elif not data.ndim == 1:
            raise NotImplementedError(
                "Multi-dimensional data columns cannot be converted to time-series"
            )
        self.check_dtime()
        if not len(data) == len(self.dtime):
            raise ValueError(
                f"Mismatch between length of data array "
                f"for variable {var_name} (length: {len(data)}) "
                f"and time array  (length: {len(self.dtime)})."
            )
        self[var_name] = s = pd.Series(data, index=self.dtime)
        return s

    def select_altitude(self, var_name, altitudes):
        """Extract variable data within certain altitude range

        Note
        ----
        Beta version

        Parameters
        ----------
        var_name : str
            name of variable for which metadata is supposed to be extracted
        altitudes : list
            altitude range in m, e.g. [0, 1000]

        Returns
        -------
        pandas. Series or xarray.DataArray
            data object within input altitude range
        """
        data = self[var_name]

        if not isrange(altitudes):
            raise NotImplementedError(
                "So far only a range (low, high) is supported for altitude extraction."
            )

        if isinstance(data, xr.DataArray):
            if not sorted(data.dims) == ["altitude", "time"]:
                raise NotImplementedError(
                    "Can only handle dataarrays that contain 2 dimensions altitude and time"
                )
            if isrange(altitudes):
                if not isinstance(altitudes, slice):
                    altitudes = slice(altitudes[0], altitudes[1])
                result = data.sel(altitude=altitudes)
                if len(result.altitude) == 0:
                    raise ValueError("no data in specified altitude range")
                return result

            raise DataExtractionError("Cannot intepret input for altitude...")

        elif isinstance(data, pd.Series) or len(self.dtime) == len(data):
            if "altitude" not in self:
                raise ValueError("Missing altitude information")
            if not isinstance(data, pd.Series):
                data = pd.Series(data, self.dtime)
            alt = self.altitude
            if not isinstance(alt, list | np.ndarray):
                raise AttributeError("need 1D altitude array")
            elif not len(alt) == len(data):
                raise DataDimensionError(
                    f"Altitude data and {var_name} data have different lengths"
                )
            mask = np.logical_and(alt >= altitudes[0], alt <= altitudes[1])
            if mask.sum() == 0:
                raise ValueError("no data in specified altitude range")
            return data[mask]

        raise DataExtractionError(
            f"Cannot extract altitudes: type of {var_name} ({type(data)}) is not supported"
        )

    def to_timeseries(self, var_name, **kwargs):
        """Get pandas.Series object for one of the data columns

        Parameters
        ----------
        var_name : str
            name of variable (e.g. "od550aer")

        Returns
        -------
        Series
            time series object

        Raises
        ------
        KeyError
            if variable key does not exist in this dictionary
        ValueError
            if length of data array does not equal the length of the time array
        """
        if var_name not in self:
            raise KeyError(f"Variable {var_name} does not exist")

        data = self[var_name]

        if isinstance(data, xr.DataArray):
            if not all([x in data.dims for x in ("time", "altitude")]):
                raise NotImplementedError(
                    "Can only handle dataarrays that "
                    "contain 2 dimensions of time "
                    "and altitude"
                )
            if "altitude" not in kwargs:
                raise ValueError(
                    "please specify altitude range via input "
                    "arg: altitude, e.g. altitude=(100,110)"
                )
            alt_info = kwargs.pop("altitude")
            data = self.select_altitude(var_name, alt_info)
            data = data.mean("altitude")
            data = data.to_series()

        if not isinstance(data, pd.Series):
            data = self._to_ts_helper(var_name)

        return data

    def plot_timeseries(self, var_name, add_overlaps=False, legend=True, tit=None, **kwargs):
        """
        Plot timeseries for variable

        Note
        ----
        If you set input arg ``add_overlaps = True`` the overlapping timeseries
        data - if it exists - will be plotted on top of the actual timeseries
        using red colour and dashed line. As the overlapping data may be
        identical with the actual data, you might want to increase the line
        width of the actual timeseries using an additional input argument
        ``lw=4``, or similar.

        Parameters
        ----------
        var_name : str
            name of variable (e.g. "od550aer")
        add_overlaps : bool
            if True and if overlapping data exists for this variable, it will
            be added to the plot.
        tit : :obj:`str`, optional
            title of plot, if None, default title is used
        **kwargs
            additional keyword args passed to matplotlib ``plot`` method

        Returns
        -------
        axes
            matplotlib.axes instance of plot

        Raises
        ------
        KeyError
            if variable key does not exist in this dictionary
        ValueError
            if length of data array does not equal the length of the time array
        """
        if "label" in kwargs:
            lbl = kwargs.pop("label")
        else:
            lbl = var_name
            try:
                ts_type = self.get_var_ts_type(var_name)
                lbl += f" ({ts_type})"
            except Exception:
                pass
        if "ax" not in kwargs:
            if "figsize" in kwargs:
                fs = kwargs.pop("figsize")
            else:
                fs = (16, 8)
            _, ax = plt.subplots(1, 1, figsize=fs)
        else:
            ax = kwargs.pop("ax")
            # keep existing title if it exists
            _tit = ax.get_title()
            if not _tit == "":
                tit = _tit

        if tit is None:
            try:
                tit = self.get_meta(force_single_value=True, quality_check=False)["station_name"]
            except Exception:
                tit = "Failed to retrieve station_name"
        s = self.to_timeseries(var_name)
        ax.plot(s, label=lbl, **kwargs)
        if add_overlaps and var_name in self.overlap:
            so = self.overlap[var_name]
            ax.plot(so, "--", lw=1, c="r", label=f"{var_name} (overlap)")

        ylabel = var_name
        try:
            if "units" in self.var_info[var_name]:
                u = self.var_info[var_name]["units"]
                if u is not None and u not in [1, "1"]:
                    ylabel += f" [{u}]"
        except Exception:
            logger.warning(f"Failed to access unit information for variable {var_name}")
        ax.set_ylabel(ylabel)
        ax.set_title(tit)
        if legend:
            ax.legend()
        return ax

    def copy(self):
        new = StationData()
        for key, val in self.items():
            cpv = deepcopy(val)
            new[key] = cpv

        return new

    def __str__(self):
        """String representation"""
        head = f"Pyaerocom {type(self).__name__}"
        s = f"\n{head}\n{len(head) * '-'}"
        arrays = ""
        series = ""

        for k, v in self.items():
            if k[0] == "_":
                continue
            if isinstance(v, dict):
                s += f"\n{k} ({type(v).__name__}):"
                if v:
                    s += dict_to_str(v, indent=2)
                else:
                    s += " <empty_dict>"
            elif isinstance(v, list):
                s += f"{k} : {list_to_shortstr(v)}"
            elif isinstance(v, np.ndarray):
                if v.ndim == 1:
                    arrays += f"{k} : {list_to_shortstr(v)}"
                else:
                    arrays += f"\n{k} (ndarray, shape {v.shape})"
                    arrays += f"\n{v}"
            elif isinstance(v, pd.Series):
                series += f"\n{k} (Series, {len(v)} items)"
            else:
                if isinstance(v, str) and v == "":
                    v = "<empty_str>"
                s += f"\n{k}: {v}"
        if arrays:
            s += "\n\nData arrays\n................."
            s += arrays
        if series:
            s += "\nPandas Series\n................."
            s += series

        return s



"""
Module containing pyaerocom custom exceptions
"""


class AeronetReadError(IOError):
    # Aeronet reading failed somehow
    pass


class CachingError(IOError):
    pass


class CacheWriteError(CachingError):
    pass


class CacheReadError(CachingError):
    pass


class ColocationError(ValueError):
    pass


class ColocationSetupError(ValueError):
    pass


class CoordinateError(ValueError):
    pass


class CoordinateNameError(CoordinateError):
    pass


class DataRetrievalError(IOError):
    pass


class DataCoverageError(ValueError):
    pass


class DataDimensionError(ValueError):
    pass


class DataIdError(ValueError):
    pass


class DataQueryError(ValueError):
    pass


class DataSourceError(ValueError):
    pass


class DataUnitError(ValueError):
    pass


class DeprecationError(AttributeError, ValueError):
    pass


class DimensionOrderError(DataDimensionError):
    pass


class DataExtractionError(ValueError):
    pass


class DataSearchError(IOError):
    pass


class EvalEntryNameError(KeyError):
    pass


class NasaAmesReadError(IOError):
    pass


class EbasFileError(ValueError):
    pass


class EEAv2FileError(ValueError):
    pass


class EntryNotAvailable(KeyError):
    pass


class InitialisationError(ValueError):
    pass


class FileConventionError(IOError):
    pass


class LongitudeConstraintError(ValueError):
    pass


class MetaDataError(AttributeError):
    pass


class ModelVarNotAvailable(IOError):
    pass


class NetworkNotSupported(NotImplementedError):
    pass


class NetworkNotImplemented(NotImplementedError):
    pass


class NetcdfError(IOError):
    pass


class NotInFileError(IOError):
    pass


class ResamplingError(ValueError):
    pass


class StationCoordinateError(CoordinateError):
    pass


class StationNotFoundError(AttributeError):
    pass


class TimeZoneError(AttributeError):
    pass


class TimeMatchError(AttributeError):
    pass


class TemporalResolutionError(ValueError):
    pass


class TemporalSamplingError(ValueError):
    pass


class UnitConversionError(ValueError):
    pass


class UnknownRegion(ValueError):
    pass


class UnresolvableTimeDefinitionError(DataDimensionError, NetcdfError):
    """Is raised if time definition in NetCDF file is wrong and cannot be corrected"""

    pass


class VarNotAvailableError(DataCoverageError):
    pass


class VariableDefinitionError(IOError):
    pass


class VariableNotFoundError(IOError):
    pass



import os
from copy import deepcopy
from typing import Literal

import pytest

from pyaerocom.aeroval import EvalSetup
from tests.fixtures.aeroval.cfg_test_exp1 import CFG
from pyaerocom.aeroval.modelmaps_helpers import CONTOUR


@pytest.fixture
def cfg_exp1(update: dict | None) -> dict:
    cfg = deepcopy(CFG)
    if update:
        cfg.update(update)

    return cfg


@pytest.fixture
def eval_setup(cfg_exp1: dict) -> EvalSetup:
    return EvalSetup.model_validate(cfg_exp1)


default_config = pytest.mark.parametrize("update", (pytest.param(None, id="default"),))


@default_config
def test_EvalSetup(cfg_exp1: dict):
    assert EvalSetup(**cfg_exp1) == EvalSetup.model_validate(cfg_exp1)


@pytest.mark.parametrize(
    "update",
    (
        pytest.param(None, id="defaults"),
        pytest.param(dict(proj_id="blah"), id="custom"),
    ),
)
def test_EvalSetup_ProjectInfo(eval_setup: EvalSetup, cfg_exp1: dict):
    assert eval_setup.proj_info.proj_id == cfg_exp1["proj_id"]


@pytest.mark.parametrize(
    "update",
    (
        pytest.param(None, id="defaults"),
        pytest.param(
            dict(
                exp_id="exp42",
                exp_descr="Hello world!",
                exp_name="Lorem Ipsum...",
                public=True,
            ),
            id="custom1",
        ),
        pytest.param(
            dict(
                exp_id="exp54",
                exp_descr="Hello world!",
                exp_name="Lorem Ipsum...",
                public=False,
            ),
            id="custom2",
        ),
    ),
)
def test_EvalSetup_ExperimentInfo(eval_setup: EvalSetup, cfg_exp1: dict):
    exp_info = eval_setup.exp_info
    assert exp_info.exp_id == cfg_exp1["exp_id"]
    assert exp_info.exp_descr == cfg_exp1["exp_descr"]
    assert exp_info.exp_name == cfg_exp1["exp_name"]
    assert exp_info.public == cfg_exp1["public"]


@pytest.mark.parametrize(
    "update",
    (
        pytest.param(None, id="defaults"),
        pytest.param(
            dict(freqs=["yearly", "monthly"], main_freq="yearly", periods=[]),
            id="custom1",
        ),
        pytest.param(
            dict(freqs=["monthly"], main_freq="monthly", periods=["2010", "2011", "2016"]),
            id="custom2",
        ),
    ),
)
def test_EvalSetup_TimeSetup(eval_setup: EvalSetup, cfg_exp1: dict):
    time_cfg = eval_setup.time_cfg
    assert time_cfg.freqs == cfg_exp1["freqs"]
    assert time_cfg.main_freq == cfg_exp1["main_freq"]
    assert time_cfg.periods == cfg_exp1["periods"]


@pytest.mark.parametrize(
    "update,start,stop",
    (
        pytest.param(
            dict(periods=["2022"]),
            "2022",
            "2023",
            id="custom1",
        ),
        pytest.param(
            dict(periods=["20220112"]),
            "2022/01/12 00:00:00",
            "2022/01/12 23:00:00",
            id="custom2",
        ),
    ),
)
def test_EvalSetup__check_time_config(
    eval_setup: EvalSetup,
    start: Literal["2022", "2022/01/12 00:00:00"],
    stop: Literal["2023", "2022/01/12 23:00:00"],
):
    eval_setup._check_time_config()
    assert str(eval_setup.colocation_opts.start) == start
    assert str(eval_setup.colocation_opts.stop) == stop


@pytest.mark.parametrize(
    "update",
    (
        pytest.param(None, id="defaults"),
        pytest.param(dict(maps_freq="yearly", maps_res_deg=10), id="custom"),
    ),
)
def test_EvalSetup_ModelMapsSetup(eval_setup: EvalSetup, cfg_exp1: dict, update: dict):
    modelmaps_opts = eval_setup.modelmaps_opts
    if update:
        assert modelmaps_opts.maps_freq == cfg_exp1["maps_freq"] == update["maps_freq"]
        assert modelmaps_opts.maps_res_deg == cfg_exp1["maps_res_deg"] == update["maps_res_deg"]
    else:  # defaults
        assert "maps_freq" not in cfg_exp1
        assert modelmaps_opts.maps_freq == "coarsest"
        assert "maps_res_deg" not in cfg_exp1
        assert modelmaps_opts.maps_res_deg == 5
        assert modelmaps_opts.plot_types == {CONTOUR}
        assert modelmaps_opts.overlay_save_format == "webp"


@pytest.mark.parametrize(
    "update",
    (
        pytest.param(None, id="defaults"),
        pytest.param(dict(obs_only=True, only_colocation=True), id="custom"),
    ),
)
def test_EvalSetup_EvalRunOptions(eval_setup: EvalSetup, cfg_exp1: dict, update: dict):
    processing_opts = eval_setup.processing_opts
    assert processing_opts.clear_existing_json == cfg_exp1["clear_existing_json"]
    assert processing_opts.only_json == cfg_exp1["only_json"]
    assert processing_opts.only_model_maps == cfg_exp1["only_model_maps"]
    if update:
        assert processing_opts.obs_only == cfg_exp1["obs_only"] == update["obs_only"]
        assert (
            processing_opts.only_colocation
            == cfg_exp1["only_colocation"]
            == update["only_colocation"]
        )
    else:  # defaults
        assert "obs_only" not in cfg_exp1
        assert processing_opts.obs_only is False
        assert "only_colocation" not in cfg_exp1
        assert processing_opts.only_colocation is False


@pytest.mark.parametrize(
    "update",
    (
        pytest.param(None, id="default"),
        pytest.param(
            dict(
                stats_min_yrs=10,
                use_diurnal=True,
                use_fairmode=True,
                weighted_stats=False,
                model_only_stats=True,
                obs_only_stats=True,
                add_trends=True,
                annual_stats_constrained=True,
            ),
            id="custom",
        ),
    ),
)
def test_EvalSetup_StatisticsSetup(eval_setup: EvalSetup, cfg_exp1: dict, update: dict):
    statistics_opts = eval_setup.statistics_opts
    if update:
        assert (
            statistics_opts.stats_min_yrs == cfg_exp1["stats_min_yrs"] == update["stats_min_yrs"]
        )
        assert statistics_opts.use_diurnal == cfg_exp1["use_diurnal"] == update["use_diurnal"]
        assert statistics_opts.use_fairmode == cfg_exp1["use_fairmode"] == update["use_fairmode"]
        assert (
            statistics_opts.weighted_stats
            == cfg_exp1["weighted_stats"]
            == update["weighted_stats"]
        )
        assert (
            statistics_opts.model_only_stats
            == cfg_exp1["model_only_stats"]
            == update["model_only_stats"]
        )
        assert (
            statistics_opts.obs_only_stats
            == cfg_exp1["obs_only_stats"]
            == update["obs_only_stats"]
        )
        assert statistics_opts.add_trends == cfg_exp1["add_trends"] == update["add_trends"]
        assert (
            statistics_opts.annual_stats_constrained
            == cfg_exp1["annual_stats_constrained"]
            == update["annual_stats_constrained"]
        )
    else:  # defaults
        assert "stats_min_yrs" not in cfg_exp1
        assert statistics_opts.stats_min_yrs == 0
        assert "use_diurnal" not in cfg_exp1
        assert statistics_opts.use_diurnal is False
        assert "use_fairmode" not in cfg_exp1
        assert statistics_opts.use_fairmode is False
        assert statistics_opts.weighted_stats == cfg_exp1["weighted_stats"] == True  # noqa: E712
        assert "model_only_stats" not in cfg_exp1
        assert statistics_opts.model_only_stats is False
        assert "obs_only_stats" not in cfg_exp1
        assert statistics_opts.obs_only_stats is False
        assert "add_trends" not in cfg_exp1
        assert statistics_opts.add_trends is False
        assert (
            statistics_opts.annual_stats_constrained  # noqa: E712
            == cfg_exp1["annual_stats_constrained"]
            == True
        )


@default_config
def test_EvalSetup_WebDisplaySetup(eval_setup: EvalSetup, cfg_exp1: dict):
    webdisp_opts = eval_setup.webdisp_opts
    # from config
    assert webdisp_opts.add_model_maps == cfg_exp1["add_model_maps"]
    assert webdisp_opts.map_zoom == cfg_exp1["map_zoom"]
    assert webdisp_opts.add_model_maps == cfg_exp1["add_model_maps"]
    # defaults
    assert "modelorder_from_config" not in cfg_exp1
    assert webdisp_opts.modelorder_from_config is True
    assert "obsorder_from_config" not in cfg_exp1
    assert webdisp_opts.obsorder_from_config is True


@default_config
def test_user_var_scale_colmap(cfg_exp1: dict):
    es1 = EvalSetup(**cfg_exp1)
    assert "deprdn" in es1.var_scale_colmap

    cfg = cfg_exp1.copy()
    cfg["var_scale_colmap_file"] = os.path.join(
        os.path.dirname(__file__), "data", "user_var_scale_colmap.ini"
    )
    esX = EvalSetup(**cfg)
    assert "deprdn" in esX.var_scale_colmap
    assert "newvar" in esX.var_scale_colmap
    assert esX.var_scale_colmap["deprdn"]["colmap"] != es1.var_scale_colmap["deprdn"]["colmap"]
    assert esX.var_scale_colmap["deprdn"]["scale"] != es1.var_scale_colmap["deprdn"]["scale"]

    cfg_fail = cfg_exp1.copy()
    cfg_fail["var_scale_colmap_file"] = os.path.join(
        os.path.dirname(__file__), "data", "buggy_var_scale_colmap.ini"
    )
    esY = EvalSetup(**cfg_fail)
    with pytest.raises(KeyError) as e:
        _ = esY.var_scale_colmap
    assert "buggy_var_scale_colmap.ini" in str(e.value)
    assert "scale" in str(e.value)


@default_config
def test_user_var_web_info(cfg_exp1: dict):
    es1 = EvalSetup(**cfg_exp1)
    assert "deprdn" in es1.var_web_info

    cfg = cfg_exp1.copy()
    cfg["var_web_info_file"] = os.path.join(
        os.path.dirname(__file__), "data", "user_var_web_info.ini"
    )
    esX = EvalSetup(**cfg)
    assert "deprdn" in esX.var_web_info
    assert "newvar" in esX.var_web_info
    assert esX.var_web_info["deprdn"][0] != es1.var_web_info["deprdn"][0]

    cfg_fail = cfg_exp1.copy()
    cfg_fail["var_web_info_file"] = os.path.join(
        os.path.dirname(__file__), "data", "buggy_var_web_info.ini"
    )
    esY = EvalSetup(**cfg_fail)
    with pytest.raises(KeyError) as e:
        _ = esY.var_web_info

    assert "buggy_var_web_info.ini" in str(e.value)
    assert "category" in str(e.value)



import logging
import os

import pooch

from pyaerocom import const

logger = logging.getLogger(__name__)

__all__ = ["download_minimal_dataset"]

#: tarfile to download
DEFAULT_TESTDATA_FILE = "testdata-minimal.tar.gz.20241120"

minimal_dataset = pooch.create(
    path=const.OUTPUTDIR,  # ~/MyPyaerocom/
    base_url="https://pyaerocom.met.no/pyaerocom-suppl",
    registry={
        "testdata-minimal.tar.gz.20220602": "md5:5d4c6455089bc93fff1fc5e2612cf439",
        "testdata-minimal.tar.gz.20220707": "md5:86fc5cb31e8123b96ef01d44fbe93c52",
        "testdata-minimal.tar.gz.20230919": "md5:7b4c55d5258da7a2b41a3a085b947fba",
        "testdata-minimal.tar.gz.20231013": "md5:f3e311c28e341a5c54d5bbba6f9849d2",
        "testdata-minimal.tar.gz.20231017": "md5:705d91e01ca7647b4c93dfe67def661f",
        "testdata-minimal.tar.gz.20231019": "md5:f8912ee83d6749fb2a9b1eda1d664ca2",
        "testdata-minimal.tar.gz.20231116": "md5:5da747f6596817295ba7affe3402b722",
        "testdata-minimal.tar.gz.20240722": "md5:7d933901c6d273d012f132c60df086cc",
        "testdata-minimal.tar.gz.20241120": "md5:4d2bc1782b1f468321817139d327e014",
    },
)


def download_minimal_dataset(
    file_name: str = DEFAULT_TESTDATA_FILE, /, extract_dir_override: str | None = None
):
    """Download test_data_file and extracts it.

    Parameters
    ----------
    file_name :
        The file name to be downloaded.
    extract_dir :
        An optional folder override to where to extract the file. By
        default files are extracted into `~/MyPyaerocom`
    """
    logger.debug(f"fetch {file_name} to {minimal_dataset.path}")

    if extract_dir_override is not None:
        extract_dir = os.path.abspath(extract_dir_override)
    else:
        extract_dir = "."

    minimal_dataset.path.joinpath("tmp").mkdir(parents=True, exist_ok=True)
    minimal_dataset.fetch(
        file_name, processor=pooch.Untar(["testdata-minimal"], extract_dir=extract_dir)
    )



import logging
import os
import sys
from datetime import timedelta
from functools import cached_property
from getpass import getuser
from pathlib import Path
from typing import Annotated, Literal
import datetime

from pyaerocom.aeroval.glob_defaults import VarWebInfo, VarWebScaleAndColormap
from pyaerocom.aeroval.obsentry import ObsEntry

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import aerovaldb
import pandas as pd
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PositiveInt,
    computed_field,
    field_serializer,
    field_validator,
)

from pyaerocom import __version__, const
from pyaerocom.aeroval.aux_io_helpers import ReadAuxHandler
from pyaerocom.aeroval.collections import ModelCollection, ObsCollection
from pyaerocom.aeroval.exceptions import ConfigError
from pyaerocom.aeroval.helpers import (
    _check_statistics_periods,
    _get_min_max_year_periods,
    check_if_year,
    BoundingBox,
)
from pyaerocom.aeroval.modelmaps_helpers import CONTOUR, OVERLAY
from pyaerocom.aeroval.json_utils import read_json, set_float_serialization_precision
from pyaerocom.colocation.colocation_setup import ColocationSetup

logger = logging.getLogger(__name__)

PLOT_TYPE_OPTIONS = ({OVERLAY}, {CONTOUR}, {OVERLAY, CONTOUR})


class OutputPaths(BaseModel):
    """
    Setup class for output paths of json files and co-located data

    This interface generates all paths required for an experiment.

    Attributes
    ----------
    proj_id : str
        project ID
    exp_id : str
        experiment ID
    json_basedir : str, Path
    avdb_resource : str, Path, None
        An aerovaldb resource identifier as expected by aerovaldb.open()[1].
        If not provided, pyaerocom will fall back to using json_basedir, for
        backwards compatibility.

        [1] https://aerovaldb.readthedocs.io/en/latest/api.html#aerovaldb.open
    """

    # Pydantic ConfigDict
    model_config = ConfigDict(arbitrary_types_allowed=True)

    _JSON_SUBDIRS: list[str] = [
        "map",
        "ts",
        "ts/diurnal",
        "scat",
        "hm",
        "hm/ts",
        "contour",
        "profiles",
    ]
    avdb_resource: Path | str | None = None

    json_basedir: Path | str = Field(
        default=os.path.join(const.OUTPUTDIR, "aeroval/data"), validate_default=True
    )
    coldata_basedir: Path | str = Field(
        default=os.path.join(const.OUTPUTDIR, "aeroval/coldata"), validate_default=True
    )

    @field_validator("json_basedir", "coldata_basedir")
    @classmethod
    def validate_basedirs(cls, v):
        if not os.path.exists(v):
            tmp = Path(v) if isinstance(v, str) else v
            tmp.mkdir(parents=True, exist_ok=True)
        return v

    proj_id: str
    exp_id: str

    def _check_init_dir(self, loc, assert_exists):
        if assert_exists and not os.path.exists(loc):
            os.makedirs(loc)
        return loc

    def get_coldata_dir(self, assert_exists=True):
        loc = os.path.join(self.coldata_basedir, self.proj_id, self.exp_id)
        return self._check_init_dir(loc, assert_exists)

    def get_json_output_dirs(self, assert_exists=True):
        out = {}
        base = os.path.join(self.json_basedir, self.proj_id, self.exp_id)
        for subdir in self._JSON_SUBDIRS:
            loc = self._check_init_dir(os.path.join(base, subdir), assert_exists)
            out[subdir] = loc
        # for cams2_83 the extra 'forecast' folder will contain the median scores if computed
        if self.proj_id == "cams2-83":
            loc = self._check_init_dir(os.path.join(base, "forecast"), assert_exists)
            out["forecast"] = loc
        return out


class ModelMapsSetup(BaseModel):
    maps_freq: Literal["hourly", "daily", "monthly", "yearly", "coarsest"] = "coarsest"
    maps_res_deg: PositiveInt = 5
    plot_types: dict[str, str | tuple[str, str]] | set[str] = {CONTOUR}
    boundaries: BoundingBox | None = None
    map_observations_only_in_right_menu: bool = False
    overlay_save_format: Literal["webp", "png"] = "webp"

    @field_validator("plot_types")
    def validate_plot_types(cls, v):
        if isinstance(v, dict):
            for m in v:
                if not isinstance(v[m], set):
                    if isinstance(v[m], str):
                        v[m] = set([v[m]])
                    else:
                        v[m] = set([*v[m]])
                if v[m] not in PLOT_TYPE_OPTIONS:
                    raise ConfigError("Model maps set up given a non-valid plot type.")
        if isinstance(v, str):
            v = set([v])
        if isinstance(v, list):  # can occur when reading a serialized config
            v = set(v)
        if v not in PLOT_TYPE_OPTIONS:
            raise ConfigError("Model maps set up given a non-valid plot type.")
        return v


class CAMS2_83Setup(BaseModel):
    use_cams2_83: bool = False


class StatisticsSetup(BaseModel, extra="allow"):
    """
    Setup options for statistical calculations

    Attributes
    ----------
    weighted_stats : bool
        if True, statistics are calculated using area weights,
        this is only relevant for gridded / gridded evaluations.
    annual_stats_constrained : bool
        if True, then only sites are considered that satisfy a potentially
        specified annual resampling constraint (see
        :attr:`pyaerocom.colocation.ColocationSetup.min_num_obs`). E.g.

        lets say you want to calculate statistics (bias,
        correlation, etc.) for monthly model / obs data for a given site and
        year. Lets further say, that there are only 8 valid months of data, and
        4 months are missing, so statistics will be calculated for that year
        based on 8 vs. 8 values. Now if
        :attr:`pyaerocom.colocation.ColocationSetup.min_num_obs` is
        specified in way that requires e.g. at least 9 valid months to
        represent the whole year, then this station will not be considered in
        case `annual_stats_constrained` is True, else it will. Defaults to
        False.
    stats_tseries_base_freq : str, optional
        The statistics Time Series display in AeroVal (under Overall Evaluation)
        is computed in intervals of a certain frequency, which is specified
        via :attr:`TimeSetup.main_freq` (defaults to monthly). That is,
        monthly colocated data is used as a basis to compute the statistics
        for each month (e.g. if you have 10 sites, then statistics will be
        computed based on 10 monthly values for each month of the timeseries,
        1 value for each site). `stats_tseries_base_freq` may be specified in
        case a higher resolution is supposed to be used as a basis to compute
        the timeseries in the resolution specified by
        :attr:`TimeSetup.main_freq` (e.g. if daily is specified here, then for
        the above example 310 values would be used - 31 for each site - to
        compute the statistics for a given month (in this case, a month with 31
        days, obviously).
    drop_stats: tuple, optional
        tuple of strings with names of statistics (as determined by keys in
        aeroval.glob_defaults.py's statistics_defaults) to not compute. For example,
        setting drop_stats = ("mb", "mab"), results in json files in hm/ts with
        entries which do not contain the mean bias and mean absolute bias,
        but the other statistics are preserved.
    stats_decimals: int, optional
        If provided, overwrites the decimals key in glod_defaults for the statistics, which has a deault of 3.
        Setting this higher of lower changes the number of decimals shown on the Aeroval webpage.
    round_floats_precision: int, optional
        Sets the precision argument for the function `pyaerocom.aaeroval.json_utils:set_float_serialization_precision`


    Parameters
    ----------
    kwargs
        any of the supported attributes, e.g.
        `StatisticsSetup(annual_stats_constrained=True)`

    """

    # Pydantic ConfigDict
    model_config = ConfigDict(protected_namespaces=())
    # StatisticsSetup attributes
    MIN_NUM: PositiveInt = 1
    weighted_stats: bool = True
    annual_stats_constrained: bool = False

    # Trends config
    add_trends: bool = False  # Adding trend calculations, only trends over the average time series over stations in a region
    avg_over_trends: bool = (
        False  # Adds calculation of avg over trends of time series of stations in region
    )
    obs_min_yrs: PositiveInt = 0  # Removes stations with less than this number of years of valid data (a year with data points in all four seasons) Should in most cases be the same as stats_min_yrs
    stats_min_yrs: PositiveInt = obs_min_yrs  # Calculates trends if number of valid years are equal or more than this. Should in most cases be the same as obs_min_yrs
    sequential_yrs: bool = False  # Whether or not the min_yrs should be sequential

    stats_tseries_base_freq: str | None = None
    forecast_evaluation: bool = False
    forecast_days: PositiveInt = 4
    use_fairmode: bool = False
    use_diurnal: bool = False
    obs_only_stats: bool = False
    model_only_stats: bool = False
    drop_stats: tuple[str, ...] = ()
    stats_decimals: int | None = None
    round_floats_precision: int | None = None

    if round_floats_precision:
        set_float_serialization_precision(round_floats_precision)


class TimeSetup(BaseModel):
    DEFAULT_FREQS: Literal["monthly", "yearly"] = "monthly"
    SEASONS: list[str] = ["all", "DJF", "MAM", "JJA", "SON"]
    main_freq: str = "monthly"
    freqs: list[str] = ["monthly", "yearly"]
    periods: list[str] = Field(default_factory=list)
    add_seasons: bool = True

    def get_seasons(self):
        """
        Get list of seasons to be analysed

        Returns :attr:`SEASONS` if :attr:`add_seasons` it True, else `[
        'all']` (only whole year).

        Returns
        -------
        list
            list of season strings for analysis

        """
        if self.add_seasons:
            return self.SEASONS
        return ["all"]

    def _get_all_period_strings(self):
        """
        Get list of all period strings for evaluation

        Returns
        -------
        list
            list of period / season strings
        """
        output = []
        for per in self.periods:
            for season in self.get_seasons():
                perstr = f"{per}-{season}"
                output.append(perstr)
        return output


class WebDisplaySetup(BaseModel):
    # Pydantic ConfigDict
    model_config = ConfigDict(protected_namespaces=())
    # WebDisplaySetup attributes
    map_zoom: Literal["World", "Europe", "xEMEP"] = "World"
    regions_how: Literal["default", "aerocom", "htap", "country"] = "default"
    map_zoom: str = "World"
    add_model_maps: bool = False
    modelorder_from_config: bool = True
    obsorder_from_config: bool = True
    var_order_menu: tuple[str, ...] = ()
    obs_order_menu: tuple[str, ...] = ()
    model_order_menu: tuple[str, ...] = ()
    hide_charts: tuple[str, ...] = ()
    hide_pages: tuple[str, ...] = ()
    ts_annotations: dict[str, str] = Field(default_factory=dict)
    pages: tuple[str, ...] = ("maps", "evaluation", "intercomp", "overall", "infos")


class EvalRunOptions(BaseModel):
    clear_existing_json: bool = True
    only_json: bool = False
    only_colocation: bool = False
    #: If True, process only maps (skip obs evaluation)
    only_model_maps: bool = False
    obs_only: bool = False


class ProjectInfo(BaseModel):
    proj_id: str


class ExperimentInfo(BaseModel):
    exp_id: str
    exp_name: str = ""
    exp_descr: str = ""
    public: bool = False
    exp_pi: str = getuser()
    pyaerocom_version: str = __version__
    creation_date: str = f"{datetime.datetime.now(datetime.timezone.utc):%Y-%m-%dT%H:%M:%S.%fZ}"


class EvalSetup(BaseModel):
    """Composite class representing a whole analysis setup

    This represents the level at which json I/O happens for configuration
    setup files.
    """

    ###########################
    ##   Pydantic ConfigDict
    ###########################
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow", protected_namespaces=())

    ########################################
    ## Regular & BaseModel-based Attributes
    ########################################

    io_aux_file: Annotated[
        Path | str, ".py file containing additional read methods for modeldata"
    ] = ""

    var_web_info_file: Annotated[Path | str, "config file containing additional variables"] = ""

    var_scale_colmap_file: Annotated[
        Path | str, "config file containing scales/ranges for variables"
    ] = ""

    _aux_funs: dict = {}

    @computed_field
    @cached_property
    def proj_info(self) -> ProjectInfo:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return ProjectInfo()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in ProjectInfo.model_fields
        }
        return ProjectInfo(**model_args)

    @computed_field
    @cached_property
    def exp_info(self) -> ExperimentInfo:
        model_args = {
            key: val for key, val in self.model_extra.items() if key in ExperimentInfo.model_fields
        }
        return ExperimentInfo(**model_args)

    @computed_field
    @cached_property
    def json_filename(self) -> str:
        """
        str: Savename of config file: cfg_<proj_id>_<exp_id>.json
        """
        return f"cfg_{self.proj_info.proj_id}_{self.exp_info.exp_id}.json"

    @cached_property
    def gridded_aux_funs(self) -> dict:
        if not bool(self._aux_funs) and os.path.exists(self.io_aux_file):
            self._import_aux_funs()
        return self._aux_funs

    @cached_property
    def var_web_info(self) -> VarWebInfo:
        return VarWebInfo(config_file=self.var_web_info_file)

    @cached_property
    def var_scale_colmap(self) -> VarWebScaleAndColormap:
        return VarWebScaleAndColormap(config_file=self.var_scale_colmap_file)

    @computed_field
    @cached_property
    def path_manager(self) -> OutputPaths:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return OutputPaths()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in OutputPaths.model_fields
        }
        return OutputPaths(**model_args)

    # Many computed_fields here have this hack to get keys from a general CFG into their appropriate respective classes
    # TODO: all these computed fields could be more easily defined if the config were
    # rigid enough to have them explicitly defined (e.g., in a TOML file), rather than dumping everything
    # into one large config dict and then dishing out the relevant parts to each class.
    @computed_field
    @cached_property
    def time_cfg(self) -> TimeSetup:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return TimeSetup()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in TimeSetup.model_fields
        }
        return TimeSetup(**model_args)

    @computed_field
    @cached_property
    def modelmaps_opts(self) -> ModelMapsSetup:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return ModelMapsSetup()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in ModelMapsSetup.model_fields
        }
        return ModelMapsSetup(**model_args)

    @computed_field
    @cached_property
    def cams2_83_cfg(self) -> CAMS2_83Setup:
        if not hasattr(self, "model_extra"):
            return CAMS2_83Setup()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in CAMS2_83Setup.model_fields
        }
        return CAMS2_83Setup(**model_args)

    @computed_field
    @cached_property
    def webdisp_opts(self) -> WebDisplaySetup:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return WebDisplaySetup()
        model_args = {
            key: val
            for key, val in self.model_extra.items()
            if key in WebDisplaySetup.model_fields
        }
        return WebDisplaySetup(**model_args)

    @computed_field
    @cached_property
    def processing_opts(self) -> EvalRunOptions:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return EvalRunOptions()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in EvalRunOptions.model_fields
        }
        return EvalRunOptions(**model_args)

    @computed_field
    @cached_property
    def statistics_opts(self) -> StatisticsSetup:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return StatisticsSetup(weighted_stats=True, annual_stats_constrained=False)
        model_args = {
            key: val
            for key, val in self.model_extra.items()
            if key in StatisticsSetup.model_fields
        }
        return StatisticsSetup(**model_args)

    @computed_field
    @cached_property
    def colocation_opts(self) -> ColocationSetup:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return ColocationSetup(save_coldata=True, keep_data=False, resample_how="mean")

        model_args = {
            key: val
            for key, val in self.model_extra.items()
            if key in ColocationSetup.model_fields
        }
        # need to pass some default values to the ColocationSetup if not provided in config
        default_dict = {
            "save_coldata": True,
            "keep_data": False,
            "resample_how": "mean",
        }
        for key in default_dict:
            if key not in model_args:
                model_args[key] = default_dict[key]

        return ColocationSetup(**model_args)

    ##################################
    ## Non-BaseModel-based attributes
    ##################################

    # These attributes require special attention b/c they're not based on Pydantic's BaseModel class.

    @computed_field
    @cached_property
    def obs_cfg(self) -> ObsCollection:
        oc = ObsCollection()
        for k, v in self.model_extra.get("obs_cfg", {}).items():
            oc.add_entry(k, v)
        return oc

    @field_serializer("obs_cfg")
    def serialize_obs_cfg(self, obs_cfg: ObsCollection):
        return obs_cfg.as_dict()

    @computed_field
    @cached_property
    def model_cfg(self) -> ModelCollection:
        mc = ModelCollection()
        for k, v in self.model_extra.get("model_cfg", {}).items():
            mc.add_entry(k, v)
        return mc

    @field_serializer("model_cfg")
    def serialize_model_cfg(self, model_cfg: ModelCollection):
        return model_cfg.as_dict()

    ###########################
    ##       Methods
    ###########################

    def get_obs_entry(self, obs_name) -> ObsEntry:
        """Returns ObsEntry instance for network obs_name"""
        return self.obs_cfg.get_entry(obs_name)

    def get_model_entry(self, model_name) -> dict:
        """Get model entry configuration

        Since the configuration files for experiments are in json format, they
        do not allow the storage of executable custom methods for model data
        reading. Instead, these can be specified in a python module that may
        be specified via :attr:`add_methods_file` and that contains a
        dictionary `FUNS` that maps the method names with the callable methods.

        As a result, this means that, by default, custom read methods for
        individual models in :attr:`model_config` do not contain the
        callable methods but only the names. This method will take care of
        handling this and will return a dictionary where potential custom
        method strings have been converted to the corresponding callable
        methods.

        Parameters
        ----------
        model_name : str
            name of model

        Returns
        -------
        dict
            Dictionary that specifies the model setup ready for the analysis
        """
        cfg = self.model_cfg.get_entry(model_name)
        cfg = cfg.prep_dict_analysis(self.gridded_aux_funs)
        return cfg

    def to_json(self, outdir: str, ignore_nan: bool = True, indent: int = 3) -> None:
        """
        Save configuration as JSON file

        Parameters
        ----------
        outdir : str
            directory where the config json file is supposed to be stored
        ignore_nan : bool
            set NaNs to Null when writing
        indent : int
            json indentation

        """
        with aerovaldb.open(
            self.path_manager.json_basedir
            if self.path_manager.avdb_resource is None
            else self.path_manager.json_basedir
        ) as db:
            with db.lock():
                db.put_config(self.json_repr(), self.proj_info.proj_id, self.exp_info.exp_id)

    @staticmethod
    def from_json(filepath: str) -> Self:
        """Load configuration from json config file"""
        settings = read_json(filepath)
        return EvalSetup(**settings)

    def json_repr(self):
        return self.model_dump()

    def _import_aux_funs(self) -> None:
        h = ReadAuxHandler(self.io_aux_file)
        self._aux_funs.update(**h.import_all())

    def _check_time_config(self) -> None:
        periods = self.time_cfg.periods
        colstart = self.colocation_opts.start
        colstop = self.colocation_opts.stop

        if len(periods) == 0:
            if colstart is None:
                raise ConfigError("Either periods or start must be set...")
            per = self.colocation_opts._period_from_start_stop()
            periods = [per]
            logger.info(
                f"periods is not set, inferred {per} from start / stop colocation settings."
            )

        self.time_cfg.periods = _check_statistics_periods(periods)
        start, stop = _get_min_max_year_periods(periods)
        start_yr = start.year
        stop_yr = stop.year
        years = check_if_year(periods)
        if not years:
            if start == stop and isinstance(start, pd.Timestamp):
                stop = start + timedelta(hours=23)
            elif isinstance(start, pd.Timestamp):
                stop = stop + timedelta(hours=23)

            if stop_yr == start_yr:
                stop_yr += 1
            if colstart is None:
                self.colocation_opts.start = start.strftime("%Y/%m/%d %H:%M:%S")
            if colstop is None:
                self.colocation_opts.stop = stop.strftime(
                    "%Y/%m/%d %H:%M:%S"
                )  # + 1  # add 1 year since we want to include stop year
        else:
            if colstart is None:
                self.colocation_opts.start = start_yr
            if colstop is None:
                self.colocation_opts.stop = (
                    stop_yr + 1
                )  # add 1 year since we want to include stop year



import warnings
from contextlib import contextmanager


@contextmanager
def ignore_warnings(category: type[Warning], *messages: str):
    """
    Ignore particular warnings with a decorator or context manager

    Parameters
    ----------
    category : subclass of Warning
        warning category to be ignored. E.g. UserWarning, DeprecationWarning.
        The default is Warning.
    *messages : str, optional
        warning messages to be ignored. E.g.
        ignore_warnings(Warning, 'Warning that can safely be ignored', 'Other warning to ignore').
        For each
        `<entry>` :func:`warnigns.filterwarnings('ignore', Warning, message=<entry>)`
        is called.

    Example
    -------
    @ignore_warnings(UserWarning)
    @ignore_warnings(DeprecationWarning)
    @ignore_warnings(Warning, 'I REALLY')
    def warn_randomly_and_add_numbers(num1, num2):
        warnings.warn(UserWarning('Harmless user warning'))
        warnings.warn(DeprecationWarning('This function is deprecated'))
        warnings.warn(Warning('I REALLY NEED TO REACH YOU'))
        return num1+num2

    """
    if not issubclass(category, Warning):
        raise ValueError("category must be a Warning subclass")

    if not messages:
        message = ""
    elif all(isinstance(msg, str) for msg in messages):
        message = "|".join(messages)
    else:
        raise ValueError("messages must be list of strings")

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=category, message=message)
        yield


def ignore_basemap_warning():  # pragma: no cover
    warnings.filterwarnings("ignore", r".*install Basemap$", UserWarning, "geonum", append=True)


def ignore_earth_radius_warning():  # pragma: no cover
    warnings.filterwarnings(
        "ignore", "Using DEFAULT_SPHERICAL_EARTH_RADIUS", UserWarning, "iris.*", append=True
    )



from __future__ import annotations

import numpy as np
import pytest

from pyaerocom.aux_var_helpers import (
    _calc_od_helper,
    calc_abs550aer,
    calc_ang4487aer,
    calc_od550aer,
    calc_od550gt1aer,
    calc_od550lt1aer,
    compute_angstrom_coeff,
    compute_od_from_angstromexp,
    vmrx_to_concx,
)


def test_calc_ang4487aer():
    data = dict(od440aer=0.2, od870aer=0.1)

    vals = calc_ang4487aer(data)
    assert vals == pytest.approx(1, abs=0.05)


@pytest.mark.parametrize(
    "data",
    [
        pytest.param(dict(), id="missing both"),
        pytest.param(dict(od440aer=0.2), id="missing od870aer"),
        pytest.param(dict(od870aer=0.1), id="missing od440aer"),
    ],
)
def test_calc_ang4487aer_error(data: dict):
    error = "Either of the two (or both) required variables (od440aer, od870aer) are not available in data"
    with pytest.raises(AttributeError) as e:
        calc_ang4487aer(data)
    assert str(e.value) == error


@pytest.mark.parametrize(
    "data,args,kwargs,result",
    [
        pytest.param(
            dict(od500aer=0.1, ang4487aer=1),
            ("od440aer",),
            dict(
                to_lambda=0.44, od_ref="od500aer", lambda_ref=0.5, use_angstrom_coeff="ang4487aer"
            ),
            0.11,
            id="1 data point",
        ),
        pytest.param(
            dict(
                od500aer=np.asarray([2, 2, np.nan, 2, np.nan, np.nan]),
                od440aer=np.ones(6) * 4,
                ang4487aer=np.zeros(6),
            ),
            ("od550aer", 0.55, "od500aer", 0.5, "od440aer", 0.44, "ang4487aer"),
            {},
            [2, 2, 4, 2, 4, 4],
            id="6 data points",
        ),
    ],
)
def test__calc_od_helper(data: dict, args: tuple, kwargs: dict, result: float | list[float]):
    aod = _calc_od_helper(data, *args, **kwargs)
    assert aod == pytest.approx(result, abs=0.05)


@pytest.mark.parametrize(
    "data,result",
    [
        pytest.param(dict(od500aer=0.1, ang4487aer=1), 0.09, id="simple"),
        pytest.param(
            dict(
                od440aer=_calc_od_helper(
                    dict(od500aer=0.1, ang4487aer=1),
                    "od440aer",
                    to_lambda=0.44,
                    od_ref="od500aer",
                    lambda_ref=0.5,
                    use_angstrom_coeff="ang4487aer",
                ),
                ang4487aer=1,
            ),
            0.09,
            id="complicated",
        ),
    ],
)
def test_calc_od550aer(data: dict, result: float):
    aod = calc_od550aer(data)
    assert aod == pytest.approx(result, abs=0.05)


def test_calc_od550gt1aer():
    data = dict(od500gt1aer=0.1, ang4487aer=1)
    aod = calc_od550gt1aer(data)
    assert aod == pytest.approx(0.09, abs=0.05)


def test_calc_od550lt1aer():
    data = dict(od500lt1aer=0.1, ang4487aer=1)
    aod = calc_od550lt1aer(data)
    assert aod == pytest.approx(0.09, abs=0.05)


@pytest.mark.parametrize(
    "od1,od2,lambda1,lambda2,result",
    [
        (0.1, 0.2, 0.6, 0.3, 1),
        (0.1, 0.2, 0.3, 0.6, -1),
    ],
)
def test_compute_angstrom_coeff(
    od1: float, od2: float, lambda1: float, lambda2: float, result: float
):
    assert compute_angstrom_coeff(od1, od2, lambda1, lambda2) == result


@pytest.mark.parametrize("angs,result", [(4, 0.77), (0, 0.1)])
def test_compute_od_from_angstromexp(angs: float, result: float):
    aod = compute_od_from_angstromexp(
        to_lambda=0.3, od_ref=0.1, lambda_ref=0.5, angstrom_coeff=angs
    )
    aod == result


@pytest.mark.parametrize(
    "inputval,p,T,vmr_unit,mmol_var,mmol_air,to_unit,desired",
    [
        (1, 101300, 293, "nmol mol-1", 48, None, "ug m-3", 1.9959),
        (1, 101300, 273, "nmol mol-1", 48, None, "ug m-3", 2.1421),
        (1, 101300, 273, "nmol mol-1", 48, None, "kg m-3", 2.1421e-9),
        (1, 101300, 273, "mol mol-1", 48, None, "kg m-3", 2.1421),
        (1, 98000, 273, "mol mol-1", 48, None, "kg m-3", 2.0724),
    ],
)
def test_vmrx_to_concx(
    inputval: float,
    p: float,
    T: float,
    vmr_unit: str,
    mmol_var: float,
    mmol_air: float | None,
    to_unit: str,
    desired: float,
):
    val = vmrx_to_concx(inputval, p, T, vmr_unit, mmol_var, mmol_air, to_unit)
    assert val == pytest.approx(desired, rel=1e-4)


@pytest.mark.parametrize(
    "data,expected_result",
    [
        pytest.param(
            dict(od500aer=0.1, ang4487aer=1, abs440aer=1, angabs4487aer=1),
            0.7999999999999999,
        ),
        pytest.param(
            dict(od500aer=0.1, ang4487aer=1, abs440aer=1, angabs4487aer=0),
            1,
        ),
    ],
)
def test_calc_abs550aer(data, expected_result):
    result = calc_abs550aer(data=data)
    assert result == pytest.approx(expected_result, rel=1e-4)


@pytest.mark.parametrize(
    "od_ref, use_angstrom_coeff",
    [
        pytest.param(None, "ang4487aer"),
        pytest.param("od500aer", None),
    ],
)
def test__calc_od_helper_raise_error(od_ref, use_angstrom_coeff):
    data = dict(od500aer=0.1, ang4487aer=1)
    with pytest.raises(AttributeError) as e:
        _calc_od_helper(
            data=data,
            var_name="od550lt1ang",
            to_lambda=0.55,
            od_ref=od_ref,
            lambda_ref=0.50,
            lambda_ref_alt=0.44,
            use_angstrom_coeff=use_angstrom_coeff,
            treshold_angstrom=1.0,
        )
    assert e.type is AttributeError



import logging
from traceback import format_exc

import iris
import iris.analysis
import iris.analysis.maths
import iris.cube
import iris.util
import numpy as np

from pyaerocom._lowlevel_helpers import merge_dicts
from pyaerocom.helpers import copy_coords_cube
from pyaerocom.molmasses import get_mmr_to_vmr_fac, get_molmass
from pyaerocom.units_helpers import get_unit_conversion_fac

logger = logging.getLogger(__name__)


CUBE_MATHS = {
    "add": iris.analysis.maths.add,
    "subtract": iris.analysis.maths.subtract,
    "divide": iris.analysis.maths.divide,
    "multiply": iris.analysis.maths.multiply,
}


def _apply_operator_cubes(cube1, cube2, operator_name, allow_coord_merge=True):
    if operator_name not in CUBE_MATHS:
        raise NotImplementedError(f"No such arithmetic cube operator implemented: {operator_name}")
    fun = CUBE_MATHS[operator_name]
    try:
        return fun(cube1, cube2)
    except ValueError as e:
        logger.warning(
            f"Could not {operator_name} cubes straight out of the box. Trying "
            f"to correct for dimension definition errors"
        )
        if "differing coordinates (time)" in repr(e):
            iris.util.unify_time_units([cube1, cube2])
            attrs = {}
            attrs.update(cube1.coord("time").attributes)
            attrs.update(cube2.coord("time").attributes)
            cube2.coord("time").attributes = attrs
            cube1.coord("time").attributes = attrs

        elif allow_coord_merge:
            copy_coords_cube(to_cube=cube2, from_cube=cube1, inplace=True)
    return fun(cube1, cube2)


def _check_input_iscube(*data_objs):
    from pyaerocom.griddeddata import GriddedData

    checked = []
    for obj in data_objs:
        if isinstance(obj, GriddedData):
            checked.append(obj.cube)
        elif isinstance(obj, iris.cube.Cube):
            checked.append(obj)
        else:
            raise ValueError(f"Invalid input: require GriddedData or Cube, got {type(obj)}")
    return checked


def _check_same_units(cube1, cube2):
    if cube1.units == cube2.units:
        return (cube1, cube2)
    from pyaerocom import const

    var1 = cube1.var_name
    u1 = cube1.units

    var2 = cube2.var_name
    u2 = cube2.units

    if var1 in const.VARS and u1 == const.VARS[var1]["units"]:
        cube2.convert_units(u1)
        return (cube1, cube2)
    elif var2 in const.VARS and u2 == const.VARS[var2]["units"]:
        cube1.convert_units(u2)
        return (cube1, cube2)

    try:
        cube2.convert_units(u1)
        return (cube1, cube2)
    except Exception:
        from pyaerocom.exceptions import UnitConversionError

        raise UnitConversionError("Failed to harmonise units")


def merge_meta_cubes(cube1, cube2):
    try:
        return merge_dicts(cube1.attributes, cube2.attributes)
    except Exception:
        logger.warning(f"WARNING: Failed to merge Cube metadata. Reason:\n{format_exc()}")
        ts_type = None
        try:
            if cube1.attributes["ts_type"] == cube2.attributes["ts_type"]:
                ts_type = cube1.attributes["ts_type"]
        except Exception:
            pass

        return {
            "NOTE": "METADATA_MERGE_FAILED",
            "meta1": cube1.attributes,
            "meta2": cube2.attributes,
            "ts_type": ts_type,
        }


def apply_rh_thresh_cubes(cube, rh_cube, rh_max=None):
    """Method that applies a low RH filter to input cube"""
    cube, rh_cube = _check_input_iscube(cube, rh_cube)
    if rh_max is None:
        from pyaerocom import const

        rh_max = const.VARS[cube.var_name]["dry_rh_max"]
    if not cube.shape == rh_cube.shape:
        raise ValueError
    mask = rh_cube.data > rh_max
    cube_out = cube.copy()
    cube_out.data[mask] = np.nan
    cube_out.attributes.update(cube.attributes)
    cube_out.attributes["rh_max"] = rh_max
    return cube_out


def add_cubes(cube1, cube2):
    """Method to add cubes from 2 gridded data objects"""
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube1, cube2 = _check_same_units(cube1, cube2)

    cube_out = _apply_operator_cubes(cube1, cube2, "add", allow_coord_merge=True)

    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out


def subtract_cubes(cube1, cube2):
    """Method to subtract 1 cube from another"""
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube1, cube2 = _check_same_units(cube1, cube2)

    cube_out = _apply_operator_cubes(cube1, cube2, "subtract", allow_coord_merge=True)

    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out


def multiply_cubes(cube1, cube2):
    """Method to multiply 2 cubes"""
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube_out = _apply_operator_cubes(cube1, cube2, "multiply", allow_coord_merge=True)

    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out


def divide_cubes(cube1, cube2):
    """Method to divide 2 cubes with each other"""
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube_out = _apply_operator_cubes(cube1, cube2, "divide", allow_coord_merge=True)

    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out


def lifetime_from_load_and_dep(load, wetdep, drydep):
    """Compute lifetime from load and wet and dry deposition"""
    raise NotImplementedError("Lifetime cannot be computed based on grid data")
    # load, wetdep, drydep = _check_input_iscube(load, wetdep, drydep)
    # deptot = _apply_operator_cubes(wetdep, drydep, "add", allow_coord_merge=True)

    # deptot.attributes.update(merge_meta_cubes(wetdep, drydep))

    # cube_out = _apply_operator_cubes(load, deptot, "divide", allow_coord_merge=True)
    # cube_out.attributes.update(merge_meta_cubes(load, deptot))
    # return cube_out


def compute_angstrom_coeff_cubes(cube1, cube2, lambda1=None, lambda2=None):
    """Compute Angstrom coefficient cube based on 2 optical densitiy cubes

    Parameters
    ----------
    cube1 : iris.cube.Cube
        AOD at wavelength 1
    cube2 : iris.cube.Cube
        AOD at wavelength 2
    lambda1 : float
        wavelength 1
    lambda 2 : float
        wavelength 2

    Returns
    -------
    Cube
        Cube containing Angstrom exponent(s)
    """
    from cf_units import Unit

    from pyaerocom.varnameinfo import VarNameInfo

    cube1, cube2 = _check_input_iscube(cube1, cube2)
    if lambda1 is None:
        lambda1 = VarNameInfo(cube1.var_name).wavelength_nm
    if lambda2 is None:
        lambda2 = VarNameInfo(cube2.var_name).wavelength_nm

    ratio = _apply_operator_cubes(cube1, cube2, "divide")
    logr = iris.analysis.maths.log(ratio)

    wvl_r = np.log(lambda1 / lambda2)
    cube_out = -1 * iris.analysis.maths.divide(logr, wvl_r)
    cube_out.units = Unit("1")
    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out


def rho_from_ts_ps(ts, ps):
    R = 287.058  # R for dry air

    rho = R * divide_cubes(ts, ps)

    rho.attributes.update(merge_meta_cubes(ts, ps))

    return rho


def mmr_from_vmr(cube):
    """
    Convvert gas volume/mole mixing ratios into mass mixing ratios.

    Parameters
    ----------
    cube : iris.cube.Cube
        A cube containing gas vmr data to be converted into mmr.

    Returns
    -------
    cube_out : iris.cube.Cube
        Cube containing mmr data.

    """
    cube = _check_input_iscube(cube)[0]

    var_name = cube.var_name
    M_dry_air = get_molmass("air_dry")
    M_variable = get_molmass(var_name)

    cube_out = (M_variable / M_dry_air) * cube
    return cube_out


def conc_from_vmr(cube, ts, ps):
    cube, ts, ps = _check_input_iscube(cube, ts, ps)

    mmr_cube = mmr_from_vmr(cube)
    rho = rho_from_ts_ps(ts, ps)

    cube_out = multiply_cubes(mmr_cube, rho)

    return cube_out


def conc_from_vmr_STP(
    cube,
):
    cube = _check_input_iscube(cube)[0]

    R = 287.058  # R for dry air

    standard_T = 293
    standard_P = 101300

    mmr_cube = mmr_from_vmr(cube)
    rho = R * standard_T / standard_P

    cube_out = rho * mmr_cube

    return cube_out


def mmr_to_vmr_cube(data):
    """
    Convert cube containing MMR data to VMR

    Parameters
    ----------
    data : iris.Cube or GriddedData
        input data object containing MMR data for a certain variable. Needs
        to have `var_name` attr. assigned and valid MMR AeroCom variable name
        (e.g. mmro3, mmrno2)

    Raises
    ------
    AttributeError
        if attr. `var_name` of input data does not start with mmr

    Returns
    -------
    iris.Cube
        cube containing mixing ratios expressed as VMR in units of nmole mole-1

    """
    from pyaerocom import GriddedData

    var = data.var_name
    if not var.startswith("mmr"):
        raise AttributeError(
            f"MMR to VMR can only be done for MMR variables (which must start "
            f"with mmr, however, the var_name of the input data is {var}"
        )

    # make sure mmr is in correct units (the function will crash if not)
    get_unit_conversion_fac(data.units, "kg kg-1")
    if isinstance(data, GriddedData):
        data = data.cube

    vmrvar = var.replace("mmr", "vmr")

    mulfac = get_mmr_to_vmr_fac(var) * 1e9  # 1e9 is to go to nmole mole-1

    cube = data * mulfac
    cube.attributes.update(data.attributes)
    cube.units = "nmole mole-1"
    cube.var_name = vmrvar
    return cube



"""
This module contains scatter plot routines for Aerocom data.
"""

import warnings
import matplotlib.pyplot as plt
import numpy as np

from pyaerocom import const
from pyaerocom._warnings import ignore_warnings
from pyaerocom.helpers import start_stop_str
from pyaerocom.mathutils import exponent
from pyaerocom.stats.stats import calculate_statistics


def plot_scatter(x_vals, y_vals, **kwargs):
    """Scatter plot

    Currently a wrapper for high-level method plot_scatter_aerocom (same module,
    see there for details)
    """
    warnings.warn(
        "matplotlib based plotting is no longer directly supported. This function may be removed in future versions.",
        DeprecationWarning,
    )

    return plot_scatter_aerocom(x_vals, y_vals, **kwargs)


def plot_scatter_aerocom(
    x_vals,
    y_vals,
    var_name=None,
    var_name_ref=None,
    x_name=None,
    y_name=None,
    start=None,
    stop=None,
    ts_type=None,
    unit=None,
    stations_ok=None,
    filter_name=None,
    lowlim_stats=None,
    highlim_stats=None,
    loglog=None,
    ax=None,
    figsize=None,
    fontsize_base=11,
    fontsize_annot=None,
    marker=None,
    color=None,
    alpha=0.5,
    **kwargs,
):
    """Method that performs a scatter plot of data in AEROCOM format

    Parameters
    ----------
    y_vals : ndarray
        1D array (or list) of model data points (y-axis)
    x_vals : ndarray
        1D array (or list) of observation data points (x-axis)
    var_name : :obj:`str`, optional
        name of variable that is plotted
    var_name_ref : :obj:`str`, optional
        name of variable of reference data
    x_name : :obj:`str`, optional
        Name of observation network
    y_name : :obj:`str`, optional
        Name / ID of model
    start : :obj:`str` or :obj`datetime` or similar
        start time of data
    stop : :obj:`str` or :obj`datetime` or similar
        stop time of data
    ts_type : str
        frequency of data
    unit : str, optional
        unit of data
    stations_ok : int, optional
        number of stations from which data were generated
    filter_name : str, optional
        name of filter
    lowlim_stats : float, optional
        lower value considered for statistical parameters
    highlim_stats : float, optional
        upper value considered for statistical parameters
    loglog : bool, optional
        plot log log scale, if None, pyaerocom default is used
    ax : Axes
        axes into which the data are to be plotted
    figsize : tuple
        size of figure (if new figure is created, ie ax is None)
    fontsize_base : int
        basic fontsize, defaults to 11
    fontsize_annot : int, optional
        fontsize used for annotations
    marker : str, optional
        marker used for data, if None, '+' is used
    color : str, optional
        color of markers, default to 'k'
    alpha : float, optional
        transparency of markers (does not apply to all marker types),
        defaults to 0.5.
    **kwargs
        additional keyword args passed to :func:`ax.plot`


    Returns
    -------
    matplotlib.axes.Axes
        plot axes
    """
    warnings.warn(
        "matplotlib based plotting is no longer directly supported. This function may be removed in future versions.",
        DeprecationWarning,
    )

    if marker is None:
        marker = "+"
    if color is None:
        color = "k"
    if isinstance(y_vals, list):
        y_vals = np.asarray(y_vals)
    if isinstance(x_vals, list):
        x_vals = np.asarray(x_vals)
    try:
        var = const.VARS[var_name]
    except Exception:
        var = const.VARS["DEFAULT"]

    try:
        var_ref = const.VARS[var_name_ref]
    except Exception:
        var_ref = const.VARS["DEFAULT"]

    if loglog is None:
        loglog = var_ref.scat_loglog

    xlim = var["scat_xlim"]
    ylim = var_ref["scat_ylim"]

    if xlim is None or ylim is None:
        low = np.min([np.nanmin(x_vals), np.nanmin(y_vals)])
        high = np.max([np.nanmax(x_vals), np.nanmax(y_vals)])

        xlim = [low, high]
        ylim = [low, high]

    if ax is None:
        if figsize is None:
            figsize = (10, 8)
        fig, ax = plt.subplots(figsize=figsize)
    if var_name is None:
        var_name = "n/d"

    statistics = calculate_statistics(y_vals, x_vals, lowlim=lowlim_stats, highlim=highlim_stats)

    if loglog:
        ax.loglog(x_vals, y_vals, ls="none", color=color, marker=marker, alpha=alpha, **kwargs)
    else:
        ax.plot(x_vals, y_vals, ls="none", color=color, marker=marker, alpha=alpha, **kwargs)

    try:
        title = start_stop_str(start, stop, ts_type)
        if ts_type is not None:
            title += f" ({ts_type})"
    except Exception:
        title = ""

    if not loglog:
        xlim[0] = 0
        ylim[0] = 0
    elif any(x[0] < 0 for x in [xlim, ylim]):
        low = np.nanmin(y_vals)
        if low != 0:
            low = 10 ** (float(exponent(abs(low)) - 1))
        xlim[0] = low
        ylim[0] = low
    with ignore_warnings(
        UserWarning,
        "Attempted to set non-positive left xlim on a log-scaled axis",
        "Attempt to set non-positive xlim on a log-scaled axis will be ignored.",
    ):
        ax.set_xlim(xlim)
    with ignore_warnings(
        UserWarning,
        "Attempted to set non-positive bottom ylim on a log-scaled axis",
        "Attempt to set non-positive ylim on a log-scaled axis will be ignored.",
    ):
        ax.set_ylim(ylim)
    xlbl = f"{x_name}"
    if var_name_ref is not None:
        xlbl += f" ({var_name_ref})"
    ax.set_xlabel(xlbl, fontsize=fontsize_base + 4)
    ax.set_ylabel(f"{y_name}", fontsize=fontsize_base + 4)

    ax.set_title(title, fontsize=fontsize_base + 4)

    ax.tick_params(labelsize=fontsize_base)

    ax.plot(xlim, ylim, "--", color="#cccccc")

    xypos = {
        "var_info": (0.01, 0.95),
        "refdata_mean": (0.01, 0.90),
        "data_mean": (0.01, 0.86),
        "nmb": (0.01, 0.82),
        "mnmb": (0.35, 0.82),
        "R": (0.01, 0.78),
        "rms": (0.35, 0.78),
        "R_kendall": (0.01, 0.74),
        "fge": (0.35, 0.74),
        "ts_type": (0.8, 0.1),
        "filter_name": (0.8, 0.06),
    }

    var_str = var_name
    _ndig = abs(exponent(statistics["refdata_mean"]) - 2)
    if unit is None:
        unit = "N/D"
    if str(unit) not in ["1", "no_unit"]:
        var_str += f" [{unit}]"

    if fontsize_annot is None:
        fontsize_annot = fontsize_base
    ax.annotate(
        f"{var_str} #: {statistics['num_valid']} # st: {stations_ok}",
        xy=xypos["var_info"],
        xycoords="axes fraction",
        fontsize=fontsize_annot + 4,
        color="red",
    )

    ax.annotate(
        f"Mean (x-data): {statistics['refdata_mean']:.{_ndig}f}; "
        f"Rng: [{np.nanmin(x_vals):.{_ndig}f}, {np.nanmax(x_vals):.{_ndig}f}]",
        xy=xypos["refdata_mean"],
        xycoords="axes fraction",
        fontsize=fontsize_annot,
        color="red",
    )

    ax.annotate(
        f"Mean (y-data): {statistics['data_mean']:.{_ndig}f}; "
        f"Rng: [{np.nanmin(y_vals):.{_ndig}f}, {np.nanmax(y_vals):.{_ndig}f}]",
        xy=xypos["data_mean"],
        xycoords="axes fraction",
        fontsize=fontsize_annot,
        color="red",
    )

    ax.annotate(
        f"NMB: {statistics['nmb']*100:.1f}%",
        xy=xypos["nmb"],
        xycoords="axes fraction",
        fontsize=fontsize_annot,
        color="red",
    )

    ax.annotate(
        f"MNMB: {statistics['mnmb']*100:.1f}%",
        xy=xypos["mnmb"],
        xycoords="axes fraction",
        fontsize=fontsize_annot,
        color="red",
    )

    ax.annotate(
        f"R (Pearson): {statistics['R']:.3f}",
        xy=xypos["R"],
        xycoords="axes fraction",
        fontsize=fontsize_annot,
        color="red",
    )

    ax.annotate(
        f"RMS: {statistics['rms']:.3f}",
        xy=xypos["rms"],
        xycoords="axes fraction",
        fontsize=fontsize_annot,
        color="red",
    )

    ax.annotate(
        f"R (Kendall): {statistics['R_kendall']:.3f}",
        xy=xypos["R_kendall"],
        xycoords="axes fraction",
        fontsize=fontsize_annot,
        color="red",
    )

    ax.annotate(
        f"FGE: {statistics['fge']:.1f}",
        xy=xypos["fge"],
        xycoords="axes fraction",
        fontsize=fontsize_annot,
        color="red",
    )
    # right lower part
    ax.annotate(
        f"{ts_type}",
        xy=xypos["ts_type"],
        xycoords="axes fraction",
        ha="center",
        fontsize=fontsize_annot,
        color="black",
    )
    ax.annotate(
        f"{filter_name}",
        xy=xypos["filter_name"],
        xycoords="axes fraction",
        ha="center",
        fontsize=fontsize_annot,
        color="black",
    )

    ax.set_aspect("equal")
    return ax



from __future__ import annotations

import logging
import re
from collections.abc import Iterable
from pathlib import Path

import xarray as xr

from pyaerocom import const
from pyaerocom.io.cnemc.reader import ReadCNEMC
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData

logger = logging.getLogger(__name__)


class ReadICOS(ReadCNEMC):
    """Class for reading ICOS (CO2) observations. HARP format so based on CNEMC reader

    Args:
        ReadCNEMC (class): Base class for this reader, based on ReadUngriddedBase

    """

    #: Mask for identifying datafiles
    _FILEMASK = "icos-*.nc"

    #: Version log of this class (for caching)
    __version__ = "0.01"

    #: Name of the dataset (OBS_ID)
    DATA_ID = const.ICOS_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: There is no global ts_type but it is specified in the data files...
    TS_TYPE = "variable"

    #: sampling frequencies found in data files
    TS_TYPES_FILE = {"hour": "hourly", "day": "daily"}

    #: field name of the start time of the measurement (in lower case)
    START_TIME_NAME = "datetime_start"

    #: filed name of the end time of the measurement (in lower case)
    END_TIME_NAME = "datetime_stop"

    #: there's no general instrument name in the data
    INSTRUMENT_NAME = "unknown"

    DATA_PRODUCT = ""

    #: functions used to convert variables that are computed
    AUX_FUNS = {}

    VAR_MAPPING = {
        "vmrco2": "CO2_volume_mixing_ratio",
        "vmrch4": "CH4_volume_mixing_ratio",
        "vmrco": "CO_volume_mixing_ratio",
    }

    FILEMASK_MAPPING = {
        "vmrco2": "icos-co2-*.nc",
        "vmrch4": "icos-ch4-*.nc",
        "vmrco": "icos-co-*.nc",
    }

    STATION_REGEX = re.compile("icos-.*-(.*)-.*-.*.nc")  # co2, ch4, co agnostic

    DEFAULT_VARS = list(VAR_MAPPING)

    DATASET_NAME = DATA_ID

    PROVIDES_VARIABLES = list(VAR_MAPPING) + list(AUX_FUNS)

    def __init__(self, data_id: str | None = None, data_dir: str | None = None):
        if data_dir is None:
            data_dir = const.OBSLOCS_UNGRIDDED[const.ICOS_NAME]

        super().__init__(data_id=data_id, data_dir=data_dir)

    def read_file(
        self, filename: str | Path, vars_to_retrieve: Iterable[str] | None = None
    ) -> UngriddedData:
        """Reads data for a single year for one component"""
        if not isinstance(filename, Path):
            filename = Path(filename)
        if not filename.is_file():
            raise ValueError(f"missing {filename}")
        return self.read(vars_to_retrieve, (filename,))

    def read(
        self,
        var_to_retrieve: str | None = None,
        files: Iterable[str | Path] | None = None,
        first_file: int | None = None,
        last_file: int | None = None,
    ) -> UngriddedData:
        if var_to_retrieve is None:
            var_to_retrieve = self.DEFAULT_VARS[0]

        if isinstance(var_to_retrieve, list):
            var_to_retrieve = var_to_retrieve[0]

        if len([var_to_retrieve]) > 1:
            raise NotImplementedError("Unnskyld, ReadICOS can only read one variable at a time...")

        if unsupported := set([var_to_retrieve]) - set(self.PROVIDES_VARIABLES):
            raise ValueError(f"Unsupported variables: {', '.join(sorted(unsupported))}")

        if files is not None and not isinstance(files, tuple):
            files = tuple(files)

        if files is not None and first_file is not None:
            files = files[first_file:]

        if files is not None and last_file is not None:
            files = files[:last_file]

        if files is None:
            # # Files will depend on the var
            stations: list[StationData] = []
            this_var_files = sorted(
                Path(self.data_dir).rglob(self.FILEMASK_MAPPING[var_to_retrieve])
            )

            for station_name, paths in self.stations(files).items():
                paths_to_read = list(set(paths) & set(this_var_files))
                if not paths_to_read:
                    continue

                logger.debug(f"Reading station {station_name}")
                ds = self._read_dataset(paths_to_read)
                ds = ds.rename({self.VAR_MAPPING[var_to_retrieve]: var_to_retrieve})
                ds = ds.assign(
                    time=self._dataset_time(ds),
                    **{name: func(ds) for name, func in self.AUX_FUNS.items()},
                )
                ds.set_coords(("latitude", "longitude", "altitude"))
                stations.append(self.to_stationdata(ds, station_name))
            return UngriddedData.from_station_data(stations)
        else:
            stations: list[StationData] = []
            this_var_files = sorted(
                Path(self.data_dir).rglob(self.FILEMASK_MAPPING[var_to_retrieve])
            )

            for station_name, paths in self.stations(files).items():
                paths_to_read = list(set(paths) & set(this_var_files))
                if not paths_to_read:
                    continue

                logger.debug(f"Reading station {station_name}")
                ds = self._read_dataset(paths)
                ds = ds.rename({self.VAR_MAPPING[var_to_retrieve]: var_to_retrieve})
                ds = ds.assign(
                    time=self._dataset_time(ds),
                    **{name: func(ds) for name, func in self.AUX_FUNS.items()},
                )
                ds.set_coords(("latitude", "longitude", "altitude"))
                stations.append(self.to_stationdata(ds, station_name))

            return UngriddedData.from_station_data(stations)

    def _read_dataset(self, paths: list[Path]) -> xr.Dataset:
        return xr.open_mfdataset(
            sorted(paths),
            concat_dim="time",
            combine="nested",
            parallel=True,
            decode_cf=True,
        )

    @classmethod
    def to_stationdata(cls, ds: xr.Dataset, station_name: str) -> StationData:
        station = StationData()
        station.data_id = cls.DATA_ID
        station.dataset_name = cls.DATASET_NAME
        station.station_id = station_name
        station.station_name = station_name
        station.latitude = float(ds["latitude"][0])
        station.longitude = float(ds["longitude"][0])
        station.altitude = float(ds["altitude"][0])

        station.station_coords = {
            "latitude": station.latitude,
            "longitude": station.longitude,
            "altitude": station.altitude,
        }

        station.ts_type = "hourly"
        station["dtime"] = ds["time"].values

        for var in ds.data_vars:
            if var not in cls.PROVIDES_VARIABLES:
                continue
            station[var] = ds[var].to_series()
            station["var_info"][var] = {"units": ds[var].units}

        return station
